from .table import _OITableHDU, _OITableHDU1, _OITableHDU2
from .table import _OIFITS1HDU, _OIFITS2HDU
from .array import _MayHaveArrayHDU,_MustHaveArrayHDU
from .target import _MustHaveTargetHDU
from .wavelength import _MustHaveWavelengthHDU
from .corr import _MayHaveCorrHDU
from .inspol import _MayHaveInspolHDU

from .. import utils as _u
from .. import coo as _coo

from astropy.io import fits as _fits
from astropy import table as _table
from astropy import units as _units
from numpy import ma as _ma
import re as _re
import numpy as _np
from astropy.time import Time as _Time
from scipy.spatial.transform import Rotation as _rotation
from astropy.coordinates import SkyCoord as _SkyCoord

class _DataHDU(_OITableHDU):
    
    _CARDS = [
        ('DATE-OBS', True, _u.is_nonempty, None, 'Date at start of observation'),
    ]
    _COLUMNS = [
        ('TARGET_ID', True, '1I', (),         _u.is_strictpos, None,  None,
            'Target ID for cross-reference'), 
        ('MJD',       True, '1D', (),         None,            None,  "d",
            'Modified Julian Day at start of observation'),
        ('INT_TIME',  True, '1D', (),         None,            None,  "s",
            'Integration time'), 
        ('FLAG',      True, 'L',  ('NWAVE',), None,            False, None,
            'Flag for bad quality'),
    ]
    
    def get_obs_type(self, name, shape='data', flatten=False):
        """

Get the type of observable.

Arguments
---------

name (str):
    Name of the observable (VISAMP, VISPHI, IVIS, RVIS, FLUXDATA, VIS2DATA,
    T3AMP, T3PHI).

shape (str):
    Shape of the returned argument (optional):
        * 'none': returns a scalar
        * 'table': returns a 1D array with the same length as the table 
            (NROWS)
        * 'data': returns an array with the same shape as the observable 
            (NROWS × NWAVE) 

Returns
-------

Type of observable (uncalibrated flux, calibrated flux, correlated flux, 
absolute, differential).

        """
        return self._resize_data('N/A', shape, flatten)

    def get_uv(self, shape='table', flatten=False):
        """

Get the (u, v) coordinates.

Arguments
---------

shape (str):
    Shape of the returned argument (optional):
        * 'table': returns a 2D array with th
            (2 × NROWS or 4 × NROWS)
        * 'data': returns an array with the same shape as the observable 
            (2 × NROWS × NWAVE or 4 × NROWS × NWAVE) 

Returns
-------

uv (array of float):
    (u, v) for 2T observables and (u1, v1, u2, v2) for 3T observables.


        """
        uv = [self._resize_data(self.data[x], shape_flatten)
                    for x in self._COORD_COLUMNS]

        return _np.array(uv)

    def _update_targetid(self, index_map):

        hdu = _np.copy(self)
        hdu.TARGET_ID = [index_map[id] for id in hdu.TARGET_ID]
        return hdu
    
    def _table_colnames(self, full_uv=False, correlations=False): 

        COORD = self._get_uvcoord_names(full_uv=full_uv)
        CORREL = []
        if correlations:
            CORREL = ['CORRNAME', 'CORRINDX']

        return [
            'TARGET', 'CHANNEL', 'REF_CHANNEL_BITFIELD',
            'EFF_WAVE', 'EFF_BAND', *COORD, 
            'observable', 'type', 'value', 'error', *CORREL,
            'INSNAME', 'ARRNAME', 'STA_CONFIG', 'MJD', 'INT_TIME'
        ]

    def _table_cols(self, full_uv=False, correlations=False):
        
        names = self._table_colnames(full_uv=full_uv, correlations=correlations)
        cols = []

        colnames = self.columns.names
        obs_names = [n for n in self.get_observable_names() if n in colnames]
        err_names = [n for n in self.get_error_names() if n in colnames]
        
        def getf(n): return self.get_field(n, 'data', True, default=0)
        def gett(n): return self.get_obs_type(n, 'data', True)
        def getc(n): return self.get_corrindx(n, 'data', True)
        def resize(x): return self._resize_data(x, 'data', True)
        def hstack(x): return _ma.hstack(x)
        
        for name in names:
            if name == 'value':
                col = hstack([getf(n) for n in obs_names])
            elif name == 'error':
                col = hstack([getf(n) for n in err_names])
            elif name == 'observable':
                col = _np.hstack([resize(n) for n in obs_names])
            elif name == 'type':
                col = _np.hstack([gett(n) for n in obs_names])
            elif name == 'CORRINDX':
                col = _np.hstack([getc(n) for n in obs_names])
            else:
                col = hstack([getf(name)] * len(obs_names))
            cols.append(col)

        return cols

    def get_field(self, name, shape='none', flatten=False, default=None):

        if name == 'INSNAME':
            return self.get_insname(shape, flatten)
        if name == 'ARRNAME':
            return self.get_arrname(shape, flatten, default)
        if name == 'CORRNAME':
            return self.get_corrname(shape, flatten, default)
        if name == 'TARGET':
            return self.get_target(shape, flatten, default)
        if name == 'EFF_WAVE':
            return self.get_wave(shape, flatten)
        if name == 'EFF_BAND':
            return self.get_band(shape, flatten)
        if name == 'CHANNEL':
            return self.get_channel(shape, flatten)
        if name == 'REF_CHANNEL_BITFIELD':
            return self.get_reference_channels(shape, flatten)
        if name == 'STA_CONFIG':
            return self.get_sta_config(shape, flatten)

        DATACOLS = self._get_spec_colnames()
        DATACOLS.remove('FLAG')

        if name not in DATACOLS:
            if hasattr(self, name):
                x = getattr(self, name)
            else:
                x = default
            x = self._resize_data(x, shape, flatten)
            x = _ma.masked_array(x, mask=not hasattr(self, name))
            return x

        mask = self.FLAG
        if hasattr(self, name):
            x = getattr(self, name)
        else:
            x = self._resize_data(default, shape)
            flag = flag | True
        x = _ma.masked_array(x, mask=mask)
        if flatten:
            x = x.ravel()
 
        return x 

    def get_reference_channels(self, shape='data', flatten=False):
        """

Get the reference channels for differential quantities.  

Arguments:
----------

shape (str):
    Shape of the returned data (optional, defaults to 'data').
    * 'data': shape is NOBS × NWAVE 
    * 'table': shape NOBS

flatten (bool, optional, default: False):
   
    Whether result should be flatten to 1D array.

Returns:
--------

A bitfield (int) indicated which wavelength channel numbers are included.  
If the quantity is not differential, 0 is returned.

        """

        visref = self._resize_data(0, 'data', flatten)
        return _ma.masked_array(visref, mask=True)

    def _to_table(self, full_uv=False, correlations=False, remove_masked=False,
            observable=None, observable_type=None, mjd_min=None, mjd_max=None,
            target=None, arrname=None, insname=None, 
            wavelmin=None, wavelmax=None):

        names = self._table_colnames(full_uv=full_uv, correlations=correlations)
        cols = self._table_cols(full_uv=full_uv, correlations=correlations)
        tab = _table.Table(cols, names=names)

        keep = _np.ones_like(tab['MJD'], dtype=bool)

        if remove_masked:
            keep *= ~tab['value'].mask 
        if mjd_min is not None:
            keep *= tab['MJD'] >= mjd_min
        if mjd_max is not None:
            keep *= tab['MJD'] <= mjd_max
        if wavelmin is not None:
            keep *= tab['EFF_WAVE'] >= wavelmin
        if wavelmax is not None:
            keep *= tab['EFF_WAVE'] <= wavelmax
        if observable is not None:
            keep *= _np.in1d(tab['observable'], _np.atleast_1d(observable))
        if observable_type is not None:
             keep *= _np.in1d(tab['type'], _np.atleast_1d(observable_type))
        if target is not None:
            keep *= _np.in1d(tab['TARGET'], _np.atleast_1d(target))
        if arrname is not None:
            keep *= _np.in1d(tab['ARRNAME'], _np.atleast_1d(arrname))
        if insname is not None:
            keep *= _np.in1d(tab['INSNAME'], _np.atleast_1d(insname))

        tab = tab[keep]

        coord_names = self._get_uvcoord_names(full_uv=full_uv)
        for x in ['INT_TIME', *coord_names]:
            tab.columns[x].format = '7.3f'
        for x in ['EFF_WAVE', 'EFF_BAND']:
            tab.columns[x].format = '7.5e'
        tab.columns['MJD'].format = '7.5f'
        for x in ['value', 'error']:
            tab.columns[x].format = '7.5g'
        
        return tab

    @classmethod
    def _get_uvcoord_names(cls, full_uv=False):
       
        if full_uv:
            return ['U1COORD', 'V1COORD', 'U2COORD', 'V2COORD']
         
        names = [c for c in cls._COLUMNS['name'] if _re.match('[UV].?COORD', c)]
        names = sorted(names, key=lambda x: x[1::-1])
        return names

    def to_table(self, remove_masked=False, **kwargs):
        """

Convert to a flat table.  It is possible to only keep data corresponding
to a given list of targets, observables, or date range. 

Arguments
---------

remove_masked (bool, default: False)
    Remove masked values.

observable (default: all are kept) 
    Observable or list of observables (e.g. VIS2DATA)

observable_type  (default: all are kept)
    Observable type or list of observable types (absolute, differential, etc.)

target  (default: all are kept)
    Target name or list of target names

insname  (default: all are kept)
    Instrument configuration name or list thereof

arrname  (default: all are kept)
    Array name or list of array names

mjd_min  (default: -inf)
    Minimum Modified Julian date

mjd_max  (default: -inf)
    Maximum Modified Julian date

wavelmin (default: 0)
    Minimum wavelength

wavelmax (default: +inf)
    Maximum wavelength

Returns:
--------

    An astropy.table.Table with one scalar observable per line.

        """
        tab = self._to_table(remove_masked=remove_masked, **kwargs)
        
        return tab

    def merge(self, *others):
        """

        Merge a set of OI extensions

        Example: hdu = hdu1.merge(hdu2, hdu3)

        """
        for ref in ['ARRNAME', 'INSNAME', 'CORRNAME', 'EXTNAME']:
            sref = self.header.get(ref, '')
            for other in others:
                oref = other.header.get(ref, '')
                if sref != oref:
                    txt = f'Cannot merge tables with {ref} = {sref} & {oref}'
                    raise RuntimeError(txt)

        return self._merge_helper(*others)

    @classmethod
    def from_data(cls, *, insname, arrname=None, corrname=None,
        date=None, mjd, int_time=0., flag=False, fits_keywords={}, **columns):

        fits_keywords = dict(arrname=arrname, insname=insname, 
                             corrname=corrname, **fits_keywords)
        columns = dict(mjd=mjd, int_time=int_time, flag=flag, **columns)

        if date is None:
            mjdobs = min(mjd - int_time / 86400. / 2)
            date = _Time(mjdobs, format='mjd').isot
    
        fits_keywords = {'DATE-OBS': date, **fits_keywords}
 
        return super().from_data(fits_keywords=fits_keywords, **columns)

    def update_uv(self):
        """

Update the (u, v) coordinates (UCOORD, VCOORD) using information of
the array and target information contained in OI_ARRAY and OI_TARGET 
tables.

Warnings
--------

(u, v) may differ from (UCOORD, VCOORD) determined by a data processing
software because

a. (UCOORD, VCOORD) can be averaged independently from MJD 
(see OIFITS standard)
b. Atmospheric refraction is dealt with approximately, while (UCOORD,
   VCOORD) may have none to full modelling of the atmosphere.


        """
        raise NotImplementedError('abstract class') 

    def get_sky_coord(self, max_distance=None):
        """
Get the full coordinates of the targets, including distance and 
motions, at the time (epoch) of observation.  Only the positions
are corrected for secular motion, the velocity and proper motions
are not computed.

Returns
-------

astropy.coordinates.SkyCoord object

        """
        # cross-reference rows of the Target HDU 
        thdu = self.get_targetHDU()
        indices = self._xmatch(refhdu=thdu, refname='TARGET_ID')
        refcoo =  thdu.get_sky_coord()

        coo = refcoo[indices]
        
        # apply epoch correction to positions
        obstime = _Time(self.MJD, format='mjd')
        coo = _coo.apply_space_motion(coo, obstime, correct_motion=False)

        if max_distance is not None:
            coo.distance[_np.isnan(coo.distance)] = max_distance

        return coo 

    def get_stauvw(self, *, refraction=False):
        """

Determine the (u, v, w) coordinates of each station with respect to
the centre of the array at the mean time of observation (MJD) using 
target and array information.

Arguments
---------

refraction (bool, optional, default: False)
    Whether the atmospheric refraction will be included in the
    calculation.

Returns
-------

uvw (float × NOBS × NSTA × 3)
    (u, v, w) coordinates for each observation and each station. NOBS
    is the number of observations (rows) in the table and NSTA the number
    of stations (NSTA = 3 for OI_T3, NSTA = 2 for OI_VIS and OI_VIS2,
    NSTA = 1 for OI_FLUX in uncalibrated mode)

Precision
---------

Sources of errors are:
1. About 7×10⁻⁵ per second of uncertainty on the effective mean time
of observation. 
2. About 5×10⁻⁵ at elevation of 45⁰ (goes as cot z), due to the 
uncertainty on temperature and pressure in the atmospheric refraction 
correction.
3. Uncertainty on the effective wavelength and relative humidity.

Discrepancies of a few centimetres for hectometric baselines have been 
observed for the VLTI.

Warning
-------

(u, v) may differ from the tabulated values in the tables (e.g. UCOORD, 
VCOORD) because data processing pipelines may average (UCOORD, VCOORD) 
independently from MJD (see OIFITS standard) and atmospheric refraction 
may be dealt with differently.

Raises
------

NotImplemented Error
    There is no associated OI_ARRAY table (possible in OIFITS v. 1)
        """   
        arrayHDU = self.get_arrayHDU() 
        waveHDU = self.get_wavelengthHDU()
        if arrayHDU is None:
            raise NotImplementedError('No OI_ARRAY is referenced')

        # If frame is SKY (e.g. pupil mask), STAXYZ are (u,v,w)
        # coordinates
        XYZ = self.get_staxyz()
        frame = arrayHDU.header.get('FRAME', 'GEOCENTRIC')
        if frame == 'SKY':
            return XYZ
 
        # enu = self.get_staenu()

        loc = arrayHDU.get_location()
        lat = loc.lat.to_value('rad')
        h = loc.height.value
        obswl = waveHDU.EFF_WAVE.mean() * _units.m
        waveHDU = self.get_wavelengthHDU()

        # Transform FK5 coordinates to apparent (atmosphere-refracted)
        # ITRS coordinates. We need to go through altaz because astropy.
        FK5 = self.get_sky_coord(max_distance=1 * _units.Mpc)
        altaz_frame = _coo.altaz_frame(loc, obswl=obswl, refraction=refraction)
        # print(f"{FK5.ra[0]=} {FK5.dec[0]=}")

        UVW = _np.empty_like(XYZ)      
 
        for i, (fk5, xyz) in enumerate(zip(FK5, XYZ)):

            altaz = fk5.transform_to(altaz_frame)
            altaz = _SkyCoord(altaz.replicate(pressure=0))
            itrs = altaz.itrs.spherical

            # (X, Y, Z)_ITRS -> (u, v, w)_SKY transform
            #
            # It is equivalent to the the (u, v, w) formula given by Eq. 2-30,
            # p. 25 in Synthesis Imaging in Radio Astronomy II, A Collection of 
            # Lectures from the Sixth NRAO/NMIMT Synthesis 
            # Imaging Summer School, G. B. Taylor, C. L. Carilli, and 
            # R. A. Perley. (eds.), ASP Conference Series, 180, 1999
            #
            # except they use a local frame centred on meridian and the 
            # hour angle, while here the frame is centred on the Greenwich
            # meridian and the longitude of the source in the ITRS is used. 
             
            lon = itrs.lon.value # target longitude in ITRS (opposite of hour
                                 # angle for an observer at Greenwich meridian)
            lat = itrs.lat.value # target declination in ITRS
            #if i == 0:
            #    print(f"{altaz.alt=} {altaz.az=}")
            #    print(f"{lon=} {lat=}") 
            wuv = _u.rotation3d(xyz, 'zy', [-lon, lat], degrees=True)
            uvw = _np.roll(wuv, -1, axis=-1)

            UVW[i] = uvw

        return UVW

    def _bin_column(self, weights, obs_name, err_name):

        data = self.data[obs_name]
        
        # spurious zero error replaced by the minimum error.  For
        # GRAVITY it only seems that's for flagged data anyway...
        error = self.data[err_name]
        error[error == 0] = error[error != 0].min()
        
        datamask = self.FLAG | _np.isnan(data)
        data = _ma.masked_array(data, mask=datamask)
        inverror2 = _ma.masked_array(error ** -2, mask=datamask)

        wsum = _ma.dot(       inverror2, weights)
        dsum = _ma.dot(data * inverror2, weights)

        new_data = dsum / wsum
        new_error = wsum ** -0.5

        nchan = new_data.shape[1]
        col = self.columns[obs_name]
        unit = col.unit
        new_fmt = f"{nchan}{col.format[-1]}"

        new_data = _fits.Column(obs_name, new_fmt, unit, array=new_data)
        new_error = _fits.Column(err_name, new_fmt, unit, array=new_error)

        return [new_data, new_error]

    def _bin_helper(self, weights):

        if weights is None:
            return self.copy()

        colnames = self.columns.names
        obs_names = [n for n in self.get_observable_names() if n in colnames]
        err_names = [n for n in self.get_error_names() if n in colnames]    
        oi_colnames = [*obs_names, *err_names, 'FLAG']
       
        new_mask = _np.dot(~self.FLAG, weights) < 1 
        
        new_cols = []
        for obs_name, err_name in zip(obs_names, err_names):
            new_cols += self._bin_column(weights, obs_name, err_name)
       
        nchan = new_mask.shape[1] 
        new_flag = _fits.Column('FLAG', f"{nchan}L", array=new_mask)
        new_cols.append(_fits.Column('FLAG', f"{nchan}L", array=new_mask))
       
        new_cols += [c for c in self.columns if c.name not in oi_colnames]
 
        binned_hdu = self.from_columns(new_cols, header=self.header)

        return binned_hdu

# OIFITS1 Table 
class _DataHDU1(
        _DataHDU, 
        _MustHaveTargetHDU,
        _MayHaveArrayHDU,
        _MustHaveWavelengthHDU,
        _OIFITS1HDU
      ):
    _COLUMNS = [('TIME', True, '1D', (), None, 0., "s",
                    'Seconds since UT at start of observation')]
        

# OIFITS1 Table rev1
class _DataHDU11( 
         _DataHDU1,
         _OITableHDU1,
      ):

    def from_data(cls, *, fits_keywords={}, mjd, time=None, **columns):

        if time is None:
            time = (mjd - int(min(mjd))) * 24 * 3600 

        return super().from_data(fits_keywords=fits_keywords,
                                    mjd=mjd, time=time, **columns)

# OIFITS2 Table
class _DataHDU2(
        _DataHDU,
        _MustHaveTargetHDU,
        _MustHaveArrayHDU,
        _MustHaveWavelengthHDU,
        _MayHaveCorrHDU,
        _MayHaveInspolHDU,
        _OIFITS2HDU, 
      ):
    def _verify(self, option='warn'):

        errors = super()._verify(option=option)

        obs_names =  self.get_observable_names()
        index = _np.ma.hstack([self.get_corrindx(n, flatten=True) 
                                                for n in obs_names])
        index = index[~index.mask]
        unique = _np.unique(index)
        if len(unique) < len(index):
            err_txt = 'repeated CORRINDX'
            self.run_option(option, err_txt, fixable=False)
        if _np.any(index <= 0):
            err_txt = 'negative or null CORRINDX'
            self.run_option(option, err_txt, fixable=False)
    
        return errors

    def get_corrindx(self, obsname, shape='none', flatten=False):

        corrindex_name = obsname + '_CORRINDEX'
        if corrindex_name not in self.columns.names:
            corrindex = _np.zeros_like(self.data[obsname], dtype=int)
            corrindex = _np.ma.masked_equal(corrindex, 0)
        else:
            corrindex = self.data[corrindex_name]
            relindex = _np.arange(self.get_nwaves())
            corrindex = corrindex[:,None] + relindex
            corrindex = _np.ma.masked_array(corrindex, mask=self.FLAG)

        if flatten:
            corrindex = corrindex.ravel()

        return corrindex


# OIFITS2 Table rev1 (new table in OIFITS2)
class _DataHDU21(
        _DataHDU2,
        _OITableHDU1
      ):
    pass

# OIFITS2 Table rev2 (table was available in OIFITS1 but was updated)
class _DataHDU22(
        _DataHDU2,
        _OITableHDU2
      ):
    
    _COLUMNS = [('TIME', True, '1D', (), _u.is_zero, 0., "s",
                    'For backwards compatibility only')]

    @classmethod
    def from_data(cls, *, fits_keywords={}, **columns):

        columns['time'] = 0.

        return super().from_data(fits_keywords=fits_keywords, **columns)




import numpy as np
from astropy.io import fits
import re
from scipy.spatial.transform import Rotation

class SequentialName(object):

    def __init__(self, s, n=1):
        s = str(s)
        if m := re.match("^(.*)_([0-9]+)$", s):
            s = m.groups()[0]
            n = int(m.groups()[1])
        self._s = s
        self._n = n

    def __str__(self):
        if self._n == 1:
            return self._s
        return f"{self._s}_{self._n}"

    def __repr__(self):
        cls = type(self).__name__
        return f"{cls}('{str(self)}')"

    def __int__(self):
        return self._n

    def __eq__(self, other):
        return self._s == other._s and self._n == other._n

    def __mod__(self, other):
        """Test if base names are equal"""
        return self._s == other._s

    def __add__(self, n):
        return type(self)(self._s, self._n + n)

    @classmethod
    def next_available(cls, used, n=None):
        if n is None:
            return cls.next_available(used, n=1)[0]
        m = set(range(1, len(used) + n + 1)) - set([int(u) for u in used])
        m = sorted(m)[0:n]
        return [cls(used[0], i) for i in m] 


def InheritConstantArray(varname, dtype=None):
    """Create a class that makes a (constant) class array inheritable, i.e.
    derived classes' array will inherit of all parent classes' elements plus
    the ones they are explicitly given.

    InheritX = InheritConstantArray('X', value=['X', 'x'])
    class A(InheritX):
        X = ['A']  # will contain ['A', 'X', 'x']

    class B(InheritX):
        X = ['B'] # will contain ['B', 'X', 'x']

    class C(A, B):
        X = ['C'] # will contain ['C', 'A', 'B', 'X', 'x']

    """
    
    basename = f"_InheritConstantArray{varname}"
    
    def init_subclass(cls):


        # it seems __class__ won't work here.  I don't understand magic.
        allbases = cls.mro()
        base = [b for b in allbases if b.__name__ == basename][0]
        super(base, cls).__init_subclass__()

        # base._X will store original cls.X (or base._X_ <- cls._X)
        uvarname = f'_{varname}' if varname[0] != '_' else f'{varname}_'
        if not hasattr(base, uvarname):
            setattr(base, uvarname,  {base: []})
        stored_values = getattr(base, uvarname)
        stored_values[cls] = cls.__dict__.get(varname, [])

        # Then we set cls.X value from base classes
        bases = [b for b in allbases if issubclass(b, base)]
        values = [i for b in bases[::-1] for i in stored_values.get(b, [])]
        if dtype is not None:
            values = np.array(values, dtype=dtype)
        setattr(cls, varname, values)
   
    # dynamic type creation 
    dct = {'__init_subclass__': init_subclass}
    base = type(basename, (object,), dct)
    
    return base

def is_frame1(s):
    return s == 'GEOCENTRIC'
def is_frame2(s):
    return s in ['GEOCENTRIC', 'SKY']
def is_fovtype2(s):
    return s in ['FVWM', 'RADIUS']

def is_veltyp(s):
    return s in ['LSR', 'BARYCENT', 'HELIOCEN', 'GEOCENTR', 'TOPOCENT']
def is_veldef(s):
    return s in ['OPTICAL', 'RADIO']

def is_zero(i):
    return i == 0

def is_oifits2(s):
    return s in ['OIFITS2']

def is_oifits1(s):
    return s in ['OIFITS', 'OIFITS1']

def is_nonempty(s):
    return type(s) is str and len(s)

def is_one(s):
    return s == 1

def is_two(s):
    return s == 2

def is_date(s):
    return is_nonemptystr(s)

def is_amptyp_rev2(s):
    return s in ['absolute', 'differential', 'correlated flux']

def is_phityp_rev2(s):
    return s in ['absolute', 'differential']

def is_num(s):
    return isinstance(s, (int, float, np.number))

def is_strictposnum(i):
    return is_num(i) and i > 0

def is_posnum(i):
    return is_num(i) and i >= 0

def is_pos(x):
    return is_num(x) and x >= 0

def is_strictpos(x):
    return is_num(x) and x > 0


def is_int(i):
    return isinstance(i, (int, np.integer))

def is_strictposint(i):
    return is_int(i) and i > 0

def is_posint(i):
    return is_int(i) and i >= 0


def is_category(s):
    return s in ['SCI', 'CAL']

def is_calstat(s):
    return s in ['U', 'C']

def is_fovtype(s):
    return s in ['RADIUS', 'FWHM']

def is_amptyp2(s):
    return s in ['absolute', 'differential', 'correlated flux']

def is_phityp2(s):
    return s in ['absolute', 'differential']


NW = 'NWAVE'

def store_default(columns, name, *, default, shape=()):
    if name not in columns:
        columns[name] = default
    if shape:
        columns[name] = _np.full(shape, columns[name])

def rotation3d(array, axes, angles, degrees=False):
    rot = Rotation.from_euler(axes, angles, degrees=degrees)
    shape = np.shape(array)
    array = rot.apply(array.reshape((-1, 3))).reshape(shape)
    return array



"""
Implementation of the OI_TARGET binary table extension
"""

from .table import _OITableHDU, _OITableHDU11, _OITableHDU22
from .. import utils as _u

import numpy as _np
from astroquery.simbad import Simbad as _Simbad
from astropy.coordinates import SkyCoord as _SkyCoord
from astropy import units as _units
from astropy.time import Time as _Time

_deg = _np.deg2rad(1)
_arcsec = _deg / 3600
_milliarcsec = _arcsec / 1000

class _MustHaveTargetHDU(_OITableHDU):

    def _get_target_field(self, name, shape='none', flatten=False,
            default=None):

        refhdu = self.get_targetHDU()
        if refhdu is None:
            val = default
        else:
            val = self._xmatch(refhdu, 'TARGET_ID', name=name)
        return self._resize_data(val, shape, flatten)

    def get_target(self, shape='none', flatten=False, default='N/A'):
        """

Get the target name associated to each row of the table

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        obj = self._container[0].header.get('OBJECT', default)
        if obj != 'MULTI':
            default = obj
        return self._get_target_field('TARGET', shape, flatten, default)

    def get_sky_coord(self):
        """
Get the full coordinates of the targets, including distance and 
motions.

Returns
-------

astropy.coordinates.SkyCoord object

        """
        coordinates = []

        refhdu = self.get_targetHDU()
        indices = self._xmatch(refhdu, 'TARGET_ID')
        rows = refhdu.data[indices]        

        # A bug in the OIFITS standard (explicit in v. 1)doesn't make a
        # distinction between epoch and equinox and impose both to be a given
        # value.  This excludes ICRS where equinox is fixed.  With the
        # recommended value 2000.0, it means FK5 J2000.0/2000.0 ≃ ICRS with ~70
        # mas accuracy.  Not interferometry-grade precision...
        equinox = _Time(rows['EQUINOX'], format='decimalyear')
        epoch = equinox
        frame = 'fk5'

        ra = rows['RAEP0'] * _units.deg
        dec = rows['DECEP0'] * _units.deg
        pm_ra = rows['PMRA'] * _units.deg / _units.yr
        pm_dec = rows['PMDEC'] * _units.deg / _units.yr
        rv = rows['SYSVEL'] * _units.m / _units.s
        plx = rows['PARALLAX'] * _units.deg / _units.yr
        plx[plx == 0] = _np.nan

        dist = 1 / plx.to_value('arcsec / yr') * _units.pc
        
        coo = _SkyCoord(
            ra=ra, dec=dec, distance=dist,
            pm_ra_cosdec=pm_ra, pm_dec=pm_dec, radial_velocity=rv,
            equinox=equinox, obstime=epoch, frame=frame,
        )


        return coo
        
    def get_equinox(self, shape='none', flatten=False):
        """

Get the equinox and epoch associated to each row of the table.

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        return self._get_target_field('EQUINOX', shape, flatten)
    def get_ra(self, shape='none', flatten=False):
        """

Get the right ascension associated to each row of the table

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        return self._get_target_field('RAEP0', shape, flatten)
    def get_dec(self, shape='none', flatten=False):
        """

Get the declination associated to each row of the table

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        return self._get_target_field('DECEP0', shape, flatten)
    def get_parallax(self, shape='none', flatten=False):
        """

Get the parallax associated to each row of the table

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        return self._get_target_field('PARALLAX', shape, flatten)
    def get_pmra(self, shape='none', flatten=False):
        """

Get the proper motion in right ascension associated to each row of 
the table

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        return self._get_target_field('PMRA', shape, flatten)
    def get_pmdec(self, shape='none', flatten=False):
        """

Get the proper motion in declination associated to each row of the table

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        return self._get_target_field('PMDEC', shape, flatten)
    def get_rv(self, shape='none', flatten=False):
        """

Get the radial velocity associated to each row of the table

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        return self._get_target_field('SYSVEL', shape, flatten)
    def get_spectype(self, shape='none', flatten=False):
        """

Get the radial velocity reference (helocientric, barycentric, etc.)
associated to each row of the table

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        return self._get_target_field('SPECTYP', shape, flatten)
    def get_category(self, shape='none', flatten=False):
        """

Get the target category (SCI/CAL) associated to each row of the table

Arguments
---------

shape (str)
    Shape of the returned array
    * None or 'table':  1D array (str × NROWS), one per row.
    * 'data': 2D array (str × NOBS × NWAVE), matching the 
        interferometric data for OI_VIS, OI_VIS2, OI_T3, and 
        OI_FLUX tables; 1D array otherwise.

flatten (bool)
    Flattens to 1D array.

        """
        return self._get_target_field('CATEGORY', shape, flatten)
    
    def get_targetHDU(self):
        """

Get the corresponding OI_TARGET HDU. 

        """
        return self._container.get_targetHDU()

    def _verify(self, option='warn'):

        errors = super()._verify(option)

        # Verify Target ID is correct (> 0) and referenced

        val = _np.unique(self.TARGET_ID)
        if not all(val >= 1):
            err_text = "'TARGET_ID' should be ≥ 1"
            err = self.run_option(option, err_text, fixable=False)
            errors.append(err)

        t = self.get_targetHDU()
        if t is not self:
            for v in val:
                if v not in t.TARGET_ID:
                    err_text = f"'TARGET_ID' not refered in TargetHDU: {v}"
                    err = self.run_option(option, err_text, fixable=False)
                    errors.append(err)

        return errors

class _TargetHDU(_MustHaveTargetHDU):
    
    _EXTNAME = 'OI_TARGET' 
    _COLUMNS = [
        ('TARGET_ID',  True, 'I',  (), _u.is_strictpos, None, None,
            'target ID for cross-reference'),
        ('RAEP0',      True, '1D',  (), None,            None, "deg",
            'right ascension at epoch'), 
        ('DECEP0',     True, '1D',  (), None,            None, "deg",
            'declination at epoch'), 
        ('EQUINOX',    True, '1E',  (), None,            None, "yr",
            'equinox and epoch'),
        ('RA_ERR',     True, '1D',  (), None,            0.,   "deg",
            'uncertainty on right ascension'),  
        ('DEC_ERR',    True, '1D',  (), None,            0.,   "deg",
            'uncertainty on declination'),
        ('SYSVEL',     True, '1D',  (), None,            None, "m/s",
            'radial velocity'), 
        ('VELTYP',     True, '8A',  (), _u.is_veltyp,    None, None,
            'reference frame for radial velocity'), 
        ('VELDEF',     True, '8A',  (), _u.is_veldef,    None, None,
            'definition for radial velocity (e.g. RADIO)'),
        ('PMRA',       True, '1D',  (), None,            0.,   "deg/yr",
            'proper motion in right ascension'), 
        ('PMDEC',      True, '1D',  (), None,            0.,   "deg/yr",
            'proper motion in declination'),
        ('PMRA_ERR',   True, '1D',  (), None,            0.,   "deg/yr",
            'uncertainty on proper motion in r.a.'), 
        ('PMDEC_ERR',  True, '1D',  (), None,            0.,   "deg/yr",
            'uncertainty on proper motion in dec.'),
        ('PARALLAX',   True, '1E',  (), None,            None, "deg",
            'parallax'), 
        ('PARA_ERR',   True, '1E',  (), None,            0.,   "deg",
            'error on parallax'),
    ]
    
    ang_dist_max = 5e-9 # approx 1 mas
    
    def _verify(self, option='warn'):

        errors = super()._verify(option)

        target_id = self.TARGET_ID
        if len(_np.unique(target_id)) == len(target_id):
            return errors

        err_text = f"Repeated TARGET_ID in {type(self).__name__}"
        err = self.run_option(option, err_text, fixable=False)
        errors.append(err)

        return errors
   
    def __add__(self, other):

        return self._merge_helper(other)

    def __mod__(self, other):
        
        return self & other
               
    def is_referred_to_by(self, other):
        return (not isinstance(other, _TargetHDU) and
                isinstance(other, _MustHaveTargetHDU))

    def merge(self, *others):
   
        container = self.get_container()
        dist_max = container._merge_target_distance
        name_match = container._merge_target_name_match
 
        def eq(x, y):
            return ((not name_match or x['TARGET'] == y['TARGET']) and
                    (dist_max > 360 or x['EQUINOX'] == y['EQUINOX']) and
                    abs(x['RAEP0'] - y['RAEP0']) <= dist_max and
                    abs(x['DECEP0'] - y['DECEP0']) <= dist_max)
                 
        return self._merge_helper(*others, id_name='TARGET_ID', equality=eq)

    @classmethod
    def from_data(cls, *, target_id=None, target, ra, dec, 
            fits_keywords={}, **columns):
        """

Build an OI_TARGET extension from data.  In the following NTARGET is
the number of targets.

Arguments
---------

version (int, optional, default: 2)
    Version of the OIFITS standard to use if it cannot be determined
    from context

target (str × NTARGET):
    Target name.
target_id (int × NTARGET, optional, defaults to 1 .. NTARGET)
    Target ID for cross-reference. 
ra (float × NTARGET)
    Right ascension in degrees
ra_err (float × NTARGET, optional, default: NaN)
    Uncertainty on the latter
dec (float × NTARGET)
    Declination in degrees
dec_err (float × NTARGET, optional, default: NaN)
    Uncertainty on the latter
parallax (float × NTARGET, optional, default: NaN)
    Parallax in degrees
para_err (float × NTARGET, optional, default: NaN)
    Uncertainty on the latter
equinox (float × NTARGET, optional, default: 2000)
    Equinox and epoch of observation in the FK5 frame.
pmra (float × NTARGET, optional, default: NaN)
    Proper movement in right ascension in degrees per yer.
pmra_err (float × NTARGET, optional, default: NaN)
    Uncertainty on the latter
pmdec (float × NTARGET, optional, default: NaN)
    Proper movement in declination in degrees
dec_err (float × NTARGET, optional, default: NaN)
    Uncertainty on the latter
sysvel (float × NTARGET, optional, default: NaN)
    Radial velocity
veltyp (str  × NTARGET, optional, default: 'BARYCENTRIC')
    Radial velocity reference frame
veldef (str  × NTARGET, optional, default: 'OPTICAL')
    Wavelength at which velocity was obtained
spectyp (str  × NTARGET, optional, default: 'UNKNOWN')
    Spectral type

Arguments in OIFITS2 only
-------------------------

category (str  × NTARGET, optional)
    Target category (SCI: science, CAL: calibrator)

Additional arguments
--------------------

Any additional keyword argument will be appended as a non-standard FITS 
column with its name prefixed with NS_ 

Warning
-------

The standard mandates EQUINOX be equal to both equinox and epoch
(see version 1, version does not modify any of it). It rules out the 
ICRS.

1. For applications needing interferometric-grade precision of the 
coordinates, take care that the FK5 frame is used, not ICRS, with
~70 milliarcseconds differences.

2. FK5 J2000.0 or ICRS coordinates with a different epoch (e.g. GAIA)
cannot be entered directly and must be converted to an FK5 system
where equinox and epoch are equal, preferably 2000.0

        """
        nrows = len(target)
        if target_id is None:
            target_id = list(range(1, nrows + 1))

        _u.store_default(columns, 'ra_err', default=_np.nan)
        _u.store_default(columns, 'dec_err', default=_np.nan)
        _u.store_default(columns, 'pmra', default=_np.nan)
        _u.store_default(columns, 'pmdec', default=_np.nan)
        _u.store_default(columns, 'pmra_err', default=_np.nan)
        _u.store_default(columns, 'pmdec_err', default=_np.nan)
        _u.store_default(columns, 'parallax', default=_np.nan)
        _u.store_default(columns, 'para_err', default=_np.nan)
        _u.store_default(columns, 'sysvel', default=_np.nan)
        _u.store_default(columns, 'veltyp', default='BARYCENTRIC')
        _u.store_default(columns, 'veldef', default='OPTICAL')
        _u.store_default(columns, 'equinox', default=2000.0)

        columns = dict(target=target, target_id=target_id, raep0=ra,
            decep0=dec, **columns) 

        return super().from_data(fits_keywords=fits_keywords, **columns)
            
    @classmethod
    def from_simbad(cls, simbad_id, *, version=2, target_id=None,
        fits_keywords={}, **columns):
        """

Create an OI_TARGET table from simbad identifiers

Arguments
---------

version (int, optional)
    Version of the OIFITS standard if it cannot be deduced from context
simbad_id (str × NTARGET)
    IDs recognised by SIMBAD
target_id (int × NTARGET, optional, default: 1 .. NTARGET)
    target ID for cross-reference 
category (str × NTARGET, optional)
    target catogory, SCI or CAL
fits_keywords (dict, optional, default: {})
    additional FITS header keywords 

Additional arguments
--------------------

Any additional keyword argument will be appended as a non-standard FITS 
column with its name prefixed with NS_ 

        """

        simbad = _Simbad()
        simbad.remove_votable_fields(*simbad.get_votable_fields()[1:])
        # OIFITS standard imposes equinox = epoch for the frame which
        # de facto excludes ICRS, we use FK5.
        simbad.add_votable_fields(
            'ra(d;A;FK5;J2000;2000)', 'dec(d;D;FK5;J2000;2000)',
            'coo_err_angle', 'coo_err_maja', 'coo_err_mina',
            'plx', 'plx_error',
            'rvz_radvel', 'rvz_wavelength',
            'pmra', 'pmdec', 'pm_err_angle', 'pm_err_maja', 'pm_err_mina',
            'sp',
        )
        tab = simbad.query_objects(simbad_id)

        # target ID must be ascii, if not, pick simbad
        main_id = tab['MAIN_ID']
        target = [s if ascii(s)[1:-1] == s else m.decode() 
                            for s, m in zip(simbad_id, main_id)]

        def tolist(x, deflt=_np.nan):
            x = [e.decode() if isinstance(e, bytes) else e for e in x.tolist()]
            x = _np.array([deflt if e in ['', None] else e for e in x])
            return x

        def ellipse_to_xy_err(a, b, theta):
            a2, b2 = a**2, b**2
            cos2, sin2 = _np.cos(theta) ** 2, _np.sin(theta) ** 2
            x = _np.sqrt(a2 * cos2 + b2 * sin2)
            y = _np.sqrt(a2 * sin2 + b2 * cos2)
            return x,y 

        # coordinates and parallaxes
        ra = tolist(tab['RA_d_A_FK5_J2000_2000'])
        dec = tolist(tab['DEC_d_D_FK5_J2000_2000'])
        a = _milliarcsec / _deg * tolist(tab['COO_ERR_MAJA'])
        b = _milliarcsec  / _deg * tolist(tab['COO_ERR_MINA'])
        theta = _deg * tolist(tab['COO_ERR_ANGLE'])
        ra_err, dec_err  = ellipse_to_xy_err(a, b, theta)
        equinox = 2000.
 
        parallax = _milliarcsec / _deg * tolist(tab['PLX_VALUE'])
        para_err = _milliarcsec / _deg * tolist(tab['PLX_ERROR'])
        
        # velocity and proper motion
        sysvel = 1e3 * tolist(tab['RVZ_RADVEL'])
        veldef = tolist(tab['RVZ_WAVELENGTH'], 'OPTICAL')
        veltyp = 'BARYCENTRIC'
        
        pmra = _milliarcsec / _deg * tolist(tab['PMRA'])
        pmdec = _milliarcsec / _deg * tolist(tab['PMDEC'])
        a = _milliarcsec / _deg * tolist(tab['PM_ERR_MAJA'])
        b = _milliarcsec / _deg * tolist(tab['PM_ERR_MINA'])
        theta = _deg * tolist(tab['PM_ERR_ANGLE'])
        pmra_err, pmdec_err = ellipse_to_xy_err(a, b, theta)
       
        # spectral type
        spectyp = tolist(tab['SP_TYPE'], 'UNKNOWN') 

        return cls.from_data(version=version, target_id=target_id,
                target=target, ra=ra, dec=dec, equinox=equinox,
                ra_err=ra_err, dec_err=dec_err,
                sysvel=sysvel, veltyp=veltyp, veldef=veldef,
                pmra=pmra, pmdec=pmdec, pmra_err=pmra_err, pmdec_err=pmdec_err,
                parallax=parallax, para_err=para_err, spectyp=spectyp,
                **columns)

def _reshape_to_table(x, nrows):
    if not _np.shape(x):
        return _np.full(x, (nrows,))
    return _np.asarray(x)

class TargetHDU1(
        _TargetHDU,
        _OITableHDU11, # OIFITS1, table rev. 1
      ):
    """

First revision of the OI_TARGET binary table, OIFITS v. 1

    """
    _COLUMNS = [
        ('TARGET',  True, '16A', (), _u.is_nonempty, None, None,
            'name of the celestial object'),
        ('SPECTYP', True, '16A', (), None,           None, None,
            'spectral type'),
    ]

class TargetHDU2(
        _TargetHDU, 
        _OITableHDU22, # OIFITS2, table rev. 2
      ):
    """

Second revision of the OI_TARGET binary table, OIFITS v. 2

    """
    _COLUMNS = [
        ('TARGET',   True,  '32A', (), _u.is_nonempty, None,  None,
            'name of the celestial object'),
        ('SPECTYP',  True,  '32A', (), None,           None,  None,
            'spectral type'),
        ('CATEGORY', False, '3A',  (), _u.is_category, 'SCI', None,
            'observation category: SCIence or CALibration'),
    ]

new_target_hdu = _TargetHDU.from_data
new_target_hdu_from_simbad = _TargetHDU.from_simbad



from .table import _OITableHDU
    
class _Referenced(_OITableHDU):
    """OI tables that are explicitely referenced to by other ones (OI_ARRAY, 
OI_WAVELENGTH, OI_CORR)."""
    
    def get_referrers(self):
        """Get all tables referring to this table."""
        
        container = self.get_container()
        if not container:
            return []

        key = self._REFERENCE_KEY
        val = self.header[key]

        refs = [h for h in container 
                    if h.header.get(key, '') == val and h is not self ] 

        return refs

    def rename(self, new_name):
        """Rename the table and updates all tables referring to it"""
        
        refkey = self._REFERENCE_KEY
        container = self.get_container()
        
        if container:

            for hdu in container:
                if hdu.header.get(refkey, None) == new_name:
                    raise RuntimeError("{refkey} = {new_name} already in use")

            for hdu in self.get_referrers():
                hdu.header[refkey] = new_name

        self.header[refkey] = new_name



from .. import utils as _u
from .. import fitsutils as _fu

from .base import _ValidHDU, _OIFITS1HDU, _OIFITS2HDU

from astropy.io import fits as _fits
import numpy as _np 
import re as _re

# All OIFITS tables will inherit a _COLUMNS structured array describing
# the columns specified in the standard
_InheritColumnDescription = _u.InheritConstantArray(
    '_COLUMNS',
    dtype=[
        ('name', 'U32'),     # TTYPEn (column name, e.g. VIS2DATA)
        ('required', bool),  # whether the column must be present
        ('format', 'U3'),    # TFORMATn (data format, e.g. 1J or 15A)
        ('shape', object),   # a tuple including 'NWAVE' if necessary
        ('test', object),    # a func that tests the validity of the value
        ('default', object), # default value, if None: no default possible
        ('unit', object),    # TUNITn. Notes:
                             # * None: no unit or empty value for TUNITn
                             # * 'any': TUNITn may be present with any value
                             # * a value: TUNITn must be present with this value
        ('comment', 'U47'),  # Comment for TTYPEn in the header
    ]
)


class _OITableHDU(
         _ValidHDU,
         _fits.BinTableHDU,
         _InheritColumnDescription,
      ):

    _REFERENCE_KEY = None

    def __init__(self, data=None, header=None, uint=False, ver=None, 
                    character_as_bytes=False):
        _fits.BinTableHDU.__init__(self, data=data, header=header, uint=uint,
                    ver=ver, character_as_bytes=character_as_bytes)

    def update(self):
        
        super().update()

        # update the comments for the standard OIFITS columns and other
        # keywords
        header = self.header
        header.set('EXTNAME', self._EXTNAME, 'OIFITS extension name')

        columns = self._get_oi_columns()
        comments = header.comments

        for index, name in enumerate(self.columns.names, start=1):
           
            ttype = f"TTYPE{index}"
            if not comments[ttype]: 
                comment = f"name of column {index}"
                if name in columns['name']:
                    column = columns[columns['name'] == name][0]
                    comment = column['comment']
                comments[ttype] = comment
           
            tform = f"TFORM{index}"
            if not comments[tform]:
                comments[tform] = f"format for {name}"
            
            tunit = f"TUNIT{index}"
            if tunit in header and not comments[tunit]: 
                comments[tunit] = f"unit of {name}"
            
            tdim = f"TDIM{index}"
            if tdim in header and not comments[tdim]:
                comments[tdim] = f"dimension of {name}"
       
        # standard v.2 makes this optional, but it's good practice
        # isn't it? 
        self.add_datasum()
        self.add_checksum()

    @classmethod
    def from_columns(cls, columns, header=None, nrows=0, fill=False,
            character_as_bytes=False):

        columns = cls._fix_column_types(columns)

        return super().from_columns(columns, header=header, nrows=nrows,
            fill=False, character_as_bytes=character_as_bytes) 

    def to_version(self, version):
        """

Transform an OIFITS table between versions of the OIFITS standard.

Arguments
---------

version (int in 1..2)
    Version

Returns
-------

Binary table of same EXTNAME conforming to OIFITS standard #version.

Warning
-------

Losses may happen because
1) some column widths are shorter in v. 1 resulting in truncation.
2) some OIFITS2 columns do not exist in OIFITS1 and will be 
    prefixed with NS_ if transforming 2 -> 1.  The 1 -> 2 
    transform will not restore the original column names.

        """

        newobj = super().to_version(version)
        if hasattr(self, '_container'):
            newobj._container = self._container
        newobj.fix_column_types()

        return newobj

    @classmethod
    def __init_subclass__(cls):

        super().__init_subclass__()
        if getattr(cls, '_EXTNAME', None) and  getattr(cls, '_OI_REVN', None): 
            _fits.hdu.base._BaseHDU.register_hdu(cls)
 
    @classmethod
    def match_header(cls, header):
        """

Internal astropy.io.fits cuisine that was inadvertently exposed by the
project.

        """
        extname = getattr(cls, '_EXTNAME', None)
        oi_revn = getattr(cls, '_OI_REVN', None)

        if extname is None or oi_revn is None:
            return NotImplementedError

        return (_fits.BinTableHDU.match_header(header) and
                header.get('EXTNAME', '') == extname and
                header.get('OI_REVN', '') == oi_revn)

    def append_lines(self, lines):
      
        merged_data = _np.hstack([self.data, lines]) 
        merged = type(self)(data=merged_data, header=self.header)

        return merged
    
    def __and__(self, other):

        return getattr(other, '_EXTNAME', '') == getattr(self, '_EXTNAME', None)

    def __mod__(self, other):

        h1, h2 = self.header, other.header

        return (self & other and
                h1.get('INSNAME', '') == h2.get('INSNAME', '') and
                h1.get('ARRNAME', '') == h2.get('ARRNAME', '') and
                h1.get('CORRNAME', '') == h2.get('CORRNAME', ''))
 
    def _xmatch(self, refhdu, refname, *, name=None, concatenate=False):
        """Helper to find target or array properties from indices"""
       
        ref_indices = getattr(refhdu, refname)
        if name is None:
            ref_values = list(range(len(ref_indices))) 
        else:
            ref_values = getattr(refhdu, name)
        xmatch = dict(zip(ref_indices, ref_values))
        indices = getattr(self, refname)
        shape = indices.shape + _np.shape(ref_values[0])
        values = _np.reshape([xmatch[i] for i in indices.flatten()], shape)
      
        if concatenate:
            values = [_np.atleast_1d(a) for a in values]
            values = ['-'.join([str(x) for x in a]) for a in values]

        return values

    def data_shape(self):

        nrows = len(self.data)
        
        # is there some wavelength data?
        colnames = self._get_spec_colnames()
        if not colnames:
            return (nrows,)

        return self.data[colnames[0]].shape

    def get_nwaves(self):
        """

Return the number of spectral channels

        """ 
        return 0 

    def rename_columns(self, **names):
        """

Rename several columns at once.

Syntax: tab.rename_columns(oldname1=newname1, ...)

        """
        self.data.dtype.names = [names.get(n, n) for n in self.data.dtype.names]
        oi_colnames = self._get_oi_columns(required=True)['name']

        for old, new in names.items():

            if old in oi_colnames:
                err_txt = 'Cannot rename a mandatory OIFITS column'
                raise RuntimeError(err_txt)
            if new in self.columns.names:
                err_txt = 'Cannot rename column to an existing one'
                raise RuntimeError(err_txt)

            self.columns[old].name = new

        self.update()

        # pyfits go look to column names in base class sometimes and it
        # doesn't always get updated.
        base = self.data
        while (isinstance(base, _fits.FITS_rec) and
                isinstance(base.base, _np.recarray)):
            base = base.base
        base.dtype = self.data.dtype
                


    def _verify(self, option='warn'):
        
        errors = super()._verify(option)

        OI_COLUMNS = self._get_oi_columns(required=False)

        # Check all required columns are present
        for c in self._get_oi_colnames(required=True):
            if c not in self.data.names:
                name = self.__class__.__name__
                err_text = f"Missing column '{c}' in {name} object"
                err = self.run_option(option, err_text, fixable=False)
                errors.append(err)
           
        # Check column types
        for coldesc in OI_COLUMNS:
            name, req, fmt, shape, test, default, unit, comment = coldesc
            if name not in self.columns.names:
                continue
            
            # check the type
            real = self.columns[name].format
            if len(fmt) == 1:
                real = real[-1]
            if fmt != real:
                err_txt = f"Column {name}: type must be {fmt} but is {real}."
                fix_txt = "Will try to fix."
                def fix(): pass
                err = self.run_option(option, err_txt, fix_txt, fix)
                errors.append(err)

            # check unit
            # * 'any': unit can be absent or present with any value
            # * None: no unit, so TUNITn may be absent or empty string
            # * other value: must be matched in TUNITn
            real = self.columns[name].unit
            spec = unit
            if (spec is None and real not in [None, ''] or
                spec not in [None, 'any'] and spec != real): 
                err_txt = f"Column {name}: unit must be {spec} but is {real}."
                fix_txt = "Fixed." 
                def fix(col=self.columns[name]): col.unit = spec
                err = self.run_option(option, err_txt, fix_txt, fix)
                errors.append(err)
            
            # check dimensionality
            dshape = self.data.dtype[name].shape
            if _u.NW in shape:
                nwave = self.get_nwaves()
                shape = tuple(nwave if d == _u.NW else d for d in shape)
            if shape != dshape:
                err_txt = f"Column {name}: shape is {dshape}, should be {shape}"
                err = self.run_option(option, err_txt, fixable=False) 
                errors.append(err)
            
            # Check values
            if test is not None:
                values = self.data[name]
                invalid = _np.array([not test(v) for v in _np.nditer(values)])
                invalid = invalid.reshape(values.shape)
                if invalid.any():
                    fixable = default is not None
                    if fixable:
                        try:
                            default = _np.array(default, dtype=dtype)
                            def fix(): values[invalid] = default
                        except:
                            fixable = False
                    if not fixable:
                        fix = None
                    val1 = values[invalid][0]
                    err_txt = f"Column '{name}' has incorrect values. "
                    err_txt += f"First encountered  '{val1}'."
                    fix_txt = f"Replaced by default value"
                    err = self.run_option(option, err_txt, fix_txt, fix, fixable)

        # Non standard columns should start have prefix_
        colnames = self.data.dtype.names
        oi_colnames = self._get_oi_colnames()
        subst = {name: f"NS_{name}" for name in colnames 
                    if name not in oi_colnames and name[0:3] != 'NS_'}
        if subst:
            nonstd = ', '.join([f"'{n}'" for n in subst.keys()])
            err_text = f"Column name(s) should start with prefix_ :'{nonstd}'."
            fix_text = "NS_ has been prefixed to column name(s)"
            def fix(h=self): h.rename_columns(**subst)
            err = self.run_option(option, err_text, fix_text, fix)
            errors.append(err)

        # Fix column types (string length, float/double)
        if 'fix' in option:
            self.fix_column_types()

        return errors

    @classmethod
    def _fix_column_types(cls, columns):

        oi_columns = cls._get_oi_columns(required=False)
        new_columns = []

        for col in columns:

            coldesc = oi_columns[oi_columns['name'] == col.name]

            if len(coldesc):
                coldesc = coldesc[0]
                try:
                    col = _fu.ascolumn(col, name=coldesc['name'],
                        unit=coldesc['unit'], format=coldesc['format'])
                except:
                    pass
            new_columns.append(col)

        return new_columns

    def fix_column_types(self):
        """

Fix the column data types if float or string widths do not conform with
the standard.

        """
        columns = self._fix_column_types(self.columns)
        fixed = any(a is not b for a, b in zip(columns, self.columns))
        
        if fixed: 
            # astropy.io.fits is a mess.
            #  both .data and .columns must be accessed.
            self.data = _fits.FITS_rec.from_columns(columns)
            for col, format in zip(self.columns, self.data.formats):
                col.format = format
            self.update()

    def __repr__(self):

        return f"<{type(self).__name__} at {hex(id(self))} ({self._diminfo()})>"
   
    def __str__(self):
        
        return f"<{type(self).__name__} ({self._diminfo()})>"
 
    # Quick access to OICOLUMNS with hdu.VI2DATA, etc.
    def __getattr__(self, s):
      
        oicolnames = self._get_oi_colnames() 
        if s in oicolnames and s in self.columns.names:
            return self.data[s][...]
        
        clsname = type(self).__name__
        err = f"'{clsname}' object has no attribute '{s}'"
        raise AttributeError(err)

    def __setattr__(self, s, v):

        oicolnames = self._get_oi_colnames()
        if s in oicolnames and s in self.columns.names:
            self.data[s] = v
        else:
            self.__dict__[s] = v

    def zero(self):

        newhdu = self.copy()
        for name in self._get_spec_colnames():
            newhdu.data[name][...] = 0
        return newhdu
    
    def _diminfo(self):
        
        ncols = len(self.columns)
        nrows, *nwave = self.data_shape()
        if len(nwave):
            return f"{ncols}C×{nrows}R×{nwave[0]}W" 
        return f"{ncols}C×{nrows}R" 

    def _resize_data(self, x, shape='none', flatten=False, copy=True):
  
        data_shape = self.data_shape()
        if len(data_shape) == 1 or x is None:
            return x
        
        if shape == 'none':
            target_shape = ()
        elif shape == 'table':
            target_shape = (len(self.data),)
        elif shape == 'data':
            target_shape = data_shape
        else:
            raise ValueError(f"shape incorrect: '{shape}'")

        x_shape = _np.shape(x)
        if target_shape:
    
            if not x_shape: # scalar
                x = _np.full(target_shape, x)
            elif x_shape[0] == target_shape[0]:
                if len(target_shape) == 2:
                    x = _np.full((target_shape[1:] + x_shape), x).swapaxes(0,1)
            else:
                msg = f"dimension mismatch: {x_shape} and {target_shape}"
                raise ValueError(msg)

            if flatten:
                target_len = _np.prod(target_shape)
                flat_shape = (target_len, *x.shape[len(target_shape):])
                x = x.reshape(flat_shape)

        elif x_shape:

            x = x.copy()

        return x


    @classmethod
    def _get_oi_columns(cls, required=False, condition=None):
        cols = cls._COLUMNS
        if required:
            cols = cols[cols['required']]
        if condition is not None:
            cols = [c for c in cols if condition(c)]
        return cols
    
    @classmethod
    def _get_oi_colnames(cls, required=False, condition=None):
        cols = cls._get_oi_columns(required, condition)
        names = [c[0] for c in cols]
        return names
    
    @classmethod
    def _get_spec_columns(cls, required=False):
        return cls._get_oi_columns(required, lambda c: c[3] == ('NWAVE',))
    
    @classmethod
    def get_error_names(cls, required=False):
        cols = cls._get_spec_colnames(required)
        return [c for c in cols if c[-3:] == 'ERR']

    @classmethod
    def get_observable_names(cls, required=False):
        cols = cls._get_spec_colnames(required)
        return [c for c in cols if c[-3:] != 'ERR' and c != 'FLAG']

    @classmethod
    def _get_spec_colnames(cls, required=False):
        cols = cls._get_spec_columns(required)
        return [c[0] for c in cols]

    def merge(self, *others):
        return self._merge_helper(*others)

    def _merge_helper(self, *others, id_name=None, equality=lambda a, b: None):
        """Merge a set of OIFITS tables of the same kind.  id_name: column
ID that must be kept unique. equality: criteria to discard redundant rows.
        """

        # Check we are merging the same type of extension 
        ext1 =  self.header['EXTNAME']
        for other in others:
            ext2 = other.header['EXTNAME']
            if ext1 != ext2:
                txt = 'cannot merge FITS extensions {ext1} with {ext2}' 
                raise TypeError(txt)

        # Determine class when merging different revisions of a table
        hdus = [self, *others]
        i_maxrevn = _np.argmax([x._OI_REVN for x in hdus])
        cls = type(hdus[i_maxrevn])

        # Merged tables will keep all columns.  Values will be zero if not 
        # defined in one of the tables
        colnames, columns = _fu.merge_columns(*hdus)

        # Merge sets of FITS rows.
        # * In each set, rows duplicating one of the previous set is eliminated
        # * For each set a map of old_id -> new_id is built to avoid
        #   duplicate IDs.
        rows = [hdu.data for hdu in hdus]
        rows, maps = _fu.merge_rows(*rows, id_name=id_name, equality=equality)
        nrows = sum(len(r) for r in rows)

        # Merge headers.  
        headers = [hdu.header for hdu in hdus]
        req_keys = cls._CARDS['name'][cls._CARDS['required']]
        header = _fu.merge_fits_headers(*headers, req_keys=req_keys)
        
        # Create an empty merged fits with the right number of rows,
        # then fill it.
        #
        # FIXME null values!
        #
        
        merged = cls.from_columns(columns, nrows=nrows, fill=1, header=header)
        
        rowmin = 0
        for i, (data, map) in enumerate(zip(rows, maps)):
        
            ndata = len(data)
            rowmax = rowmin + ndata

            for name in data.names:
                if map and name == id_name:
                    values = [map.get(i, i) for i in data[name]]
                else:
                    values = data[name]
                merged.data[name][rowmin:rowmax][...] = values 
            rowmin += ndata

        # Update in HDUs refering to other
        for hdu, map in zip(hdus, maps):
            if map and (container := getattr(hdu, '_container', None)):
                for h in container.get_OITableHDUs():
                    if h.refers_to(hdu): 
                        field = h.data[id_name]
                        for old, new in map.items():
                            field[field == old] = new
        
        return merged 

    def refers_to(self, other):
        """Whether an OIFITS table makes a reference to another one via
        ARRNAME, INSNAME, or CORRNAME"""
        return other.is_referred_to_by(self)
    
    def is_referred_to_by(self, other):
        """Whether an OIFITS table is a reference for another table via
        ARRNAME, INSNAME, or CORRNAME"""
        return False

    def __eq__(self, other):

        if not isinstance(other, _OITableHDU):
            return False

        header1, header2 = self.header, other.header
        data1, data2 = self.data, other.data

        return (header1['EXTNAME'] == header2['EXTNAME'] and
                header1.get('INSNAME', '') == header2.get('INSNAME', '') and
                header1.get('ARRNAME', '') == header2.get('ARRNAME', '') and
                header1.get('CORRNAME', '') == header2.get('CORRNAME', '') and
                len(data1) == len(data2) and 
                data1.dtype == data2.dtype and
                (data1 == data2).all())

    @classmethod
    def _get_column_shape(cls, **columns):
        
        for col in cls.get_oi_columns(required=True):
            if ((name := col['name'] in columns) and
                (shape := col['shape']) and
                (value := columns[name]) is not None):
                shape = _np.shape(value)
                return shape

        raise RuntimeError('Cannot determine table shape from input')

    @classmethod
    def _guess_shape(cls, columns):

        obs_names = cls.get_observable_names()
        if len(obs_names):
            for obs_name in obs_names:
                if obs_name in columns:
                    shape = _np.shape(columns[obs_name])
                    return shape
            raise RuntimeError('cannot guess shape from data')
        
        oi_columns = cls._get_oi_columns()
        for name in oi_columns['name']:
            if name in columns:
                shape = _np.shape(columns[name])
                return shape
        raise RuntimeError('cannot guess shape from data')
         
    @classmethod
    def from_data(cls, *, version=None, fits_keywords={}, **columns):

        if not hasattr(cls, '_OI_VER'):
            if version is None:
                version = getattr(cls, '_OI_VER', 2)
            cls = cls.get_class(version=version)

        # FITS keywords and column names are upper case 
        fits_keywords = {k.upper(): v for k, v in fits_keywords.items()
                                                    if v is not None}
        columns = {k.upper(): v for k, v in columns.items()}
       
        # prefix non-standard columns
        oi_colnames = cls._get_oi_columns(required=False)['name']
        columns = {n if n in oi_colnames else f"NS_{n}": v
                    for n, v in columns.items()}
 
        # Header
        header = _fits.Header()
        for card in cls._CARDS:
            name = card['name']
            comment = card['comment']
            if name in fits_keywords:
                header.set(name, fits_keywords[name], comment)
                del fits_keywords[name]
            elif card['required']:
                if card['default'] is not None:
                    header.set(name, card['default'], comment)
        for name, value in fits_keywords.items():
            if not isinstance(value, (tuple, list)):
                value = [value]
            header.set(name, *value)
    
        # Guess shape
        shape = cls._guess_shape(columns)
        nrows = shape[0]
        
        def full(s, x):
            if _np.ndim(x) > len(s):
                return _np.asarray(x)
            return _np.full(s, x)

        # Guess errors from data    
        obs_names = cls.get_observable_names() 
        err_names = cls.get_error_names()
        for obs, err in zip(obs_names, err_names):
            if obs in columns:
                if err not in columns:
                    columns[err] = 0. 
        fcols = []

            # official OIFITS columns
        for col in cls._get_oi_columns(required=False):
            name = col['name']
            if name not in columns:
                if col['required']:
                    columns[name] = col['default']
                else:
                    continue
            if name in [*obs_names, *err_names, 'FLAG']:
                new_shape = shape
            else:
                new_shape = (nrows,)
            array = full(new_shape, columns[name]) 
            del columns[name]
            fcol = _fu.ascolumn(array, format=col['format'], unit=col['unit'],
                name=name)
            fcols.append(fcol)

            # additional columns
        for name, array in columns.items():
            fcol = _fu.ascolumn(array, name=name)
            fcols.append(fcol)
        
        tab = super().from_columns(fcols, header=header)
        return tab
    

class _OITableHDU1(_OITableHDU):
    _OI_REVN = 1
    _CARDS = [('OI_REVN', True, _u.is_one, 1, 
        '1st revision of this table in OIFITS format')]
    
class _OITableHDU2(_OITableHDU):
    _OI_REVN = 2
    _CARDS = [('OI_REVN', True, _u.is_two, 2, 
        '2nd revision of this table in OIFITS format')]

class _OITableHDU11(
        _OIFITS1HDU,
        _OITableHDU1
      ):
    pass 

class _OITableHDU21(
        _OIFITS2HDU,
        _OITableHDU1
      ):
    pass

class _OITableHDU22(
        _OIFITS2HDU,
        _OITableHDU2
      ):
    pass



from .data import _DataHDU
from .. import utils as _u

class _T2HDU(_DataHDU):
    
    _COLUMNS = [
        ('STA_INDEX', True, '2I', (2,), _u.is_strictpos, None, None,
            'station indices in matching OI_ARRAY table'),
        ('UCOORD',    True, '1D', (),   None,            None, "m",
            'u coordinate'),
        ('VCOORD',    True, '1D', (),   None,            None, "m",
            'v coordinate'),
    ]
    
    def __getattr__(self, name):

        if name == 'U1COORD':
            return self.data['UCOORD']
        if name == 'V1COORD':
            return self.data['VCOORD']
        
        return super().__getattr__(name)

    @classmethod
    def from_data(cls, *, ucoord, vcoord, fits_keywords={}, **columns):
        
        return super().from_data(fits_keywords={}, ucoord=ucoord, vcoord=vcoord,
                **columns)

    def get_uvw(self, refraction=False):
        """

Get the (u, v, w) coordinates of the baseline

Arguments
---------

refraction (bool, optional, default: False)
    Whether refraction is taken into account

Returns
-------

(u, v, w)
    Projected baseline in metres

Precision
---------
    A few centimetres for hectometric baselines, mainly due to the uncertainty
    on the time-averaging.
   
        """
        uvw = self.get_stauvw(refraction=refraction)
        uvw = uvw[:,1,:] - uvw[:,0,:]
        u, v, w = uvw.T

        return u, v, w

    def update_uv(self):

        if self.get_arrayHDU() is None:
            return

        u, v, w = self.get_uvw()
        self.UCOORD = u
        self.VCOORD = v



from astropy.io import fits
import numpy as np
import re 
from collections import OrderedDict 

column_dtype = {
    'B': 'B', 'h': 'I', 'i': 'J', 'l': 'K', 
    'f': 'E', 'd': 'D', 'F': 'C', 'D': 'M',
    '?': 'L', 'b': 'L', 'U': 'A', 'S': 'A'
}

def ascolumn(data=None, *, name=None, format=None, unit=None, dim=None,
             null=None, bzero=None, bscale=None, **kwarg):
    """
Return data as a FITS column.

Keyword arguments
-----------------
data:
        Data. Either an array or a FITS column.  If data is None, the format
        must be given.

name (str):
        Name of the column

unit (str):
        Unit

dim (tuple):
        Dimension of column cells.

format (str):
        Column format. If absent or incomplete (letter without dimension)
        it will be completed using data shape and type.

null (int):
        Null value for integer types.

    """

    if unit in ['', 'any']:
        unit = None
     
    # If a FITS column is given, look if we need to transform it

    if isinstance(data, fits.Column):
        
        # Look if it matches specifications
        if ((name is None or data.name == name) and
            (unit is None or data.unit == unit) and
            (null is None or data.null == null) and
            (bzero is None or data.bzero == bscale) and
            (bscale is None or data.bscale == bscale) and
            (dim is None or data.dim == dim)):

            # For format a few cases are possible
            # 1. no format is specified
            # 2. format contains only one letter, no data size
            # 3. format contains data size
            if (format is None or
                len(format) == 1 and data.format[-1] == format or
                data.format == format):
                return data

        # we revert to array and will try to reshape column
        array = data.array
        if name is None:
            name = data.name
        if unit is None:
            unit = data.unit
        if null is None:
            null = data.null
        if dim is None:
            dim = data.dim
    
    else:
        array = data
   
    # If array is None
    if array is None:
        col = fits.Column(name=name, unit=unit, format=format, null=null,
                dim=dim)
        return col

    # Shape is determined from the array 
    array = np.asarray(array)
    shape = array.shape
    size = int(np.prod(shape[1:]))
   
    # Determine TFORM keyword 
    if format is None:
        dtype = array.dtype.char
        format = 'A' if dtype in 'SU' else column_dtype(dtype) 
    
    if len(format) == 1:
        len_ = size
        if format[-1] == 'A':
            array = np.asarray(array, dtype=str)
            len_ *= array.itemsize // array.dtype.alignment
        format = f"{len_}{format}"
    else:
        len_ = size
        if format[-1] == 'A':
            len_ *= int(format[:-1]) // size
        if len_ != int(format[:-1]):
            print(array)
            shape = array.shape[1:]
            txt = f"Shape {shape} does not match format {format} for {name}"
            raise RuntimeError(txt)

    # Determine TDIM keyword. 
    tdim = None
    if format[-1] == 'A' and len(shape) > 1:
        tdim = (len_, *shape[1:])
    elif len(shape) > 2:
        tdim = shape[1:]

    col = fits.Column(array=array, format=format, name=name, 
            unit=unit, dim=dim, null=null)

    return col

def merge_columns(*hdus):

    # Find all column names keeping order of appearance in hdus
    colnames = [col.name for hdu in hdus for col in hdu.columns]
    colnames = list(OrderedDict.fromkeys(colnames))

    # for each column, find all HDUs it is included and keep the
    # columns with largest data size.  
    columns = []
    for name in colnames:
        col = None
        for hdu in hdus:
            if name in hdu.columns.names:
                new_col = hdu.columns[name]
                if col is None:
                    col = new_col
                elif col.array.itemsize < new_col.array.itemsize:
                    col = new_col
        columns.append(col)

    return colnames, columns

def merge_rows(*rows, id_name=None, equality=lambda x,y: x==y):

    if id_name is None:
        return rows, [{}] * len(rows)

    # New unused IDs (in a sequence, avoiding the ones in either
    rows1 = rows[0]
    id1 = rows1[id_name].tolist()
    kept = [rows1]
    maps = [{}]

    for rows2 in rows[1:]:

        id2 = rows2[id_name]
        candidate_id2 = set(range(1, 1 + len(id1) + len(id2)))
        candidate_id2 -= set([*id1, *id2])
        candidate_id2 = sorted(list(candidate_id2))

        index_map = {}
        kept_lines = []
        for j, row2 in enumerate(rows2):

            equal = [equality(row2, row1) for rows1 in kept for row1 in rows1]
            index_equal = np.argwhere(equal)
            if len(index_equal):
                index_map[id2[j]] = id1[index_equal[0,0]]
                continue

            kept_lines.append(j)
            value = candidate_id2.pop(0) if id2[j] in id1 else id2[j]
            index_map[id2[j]] = value
            id1.append(value)

        kept2 = rows2[kept_lines]
        index_map = {o: n for o, n in index_map.items() if o != n}

        kept.append(kept2)
        maps.append(index_map)

    return kept, maps

def merge_fits_headers(*headers, req_keys=[]):

    header = headers[0].copy()

    for h in headers[1:]:
        for card in h.cards:
            (name, value, comment) = card
            if name in ['', 'HISTORY', 'COMMENT']:
                header.append(card)
            elif name not in header:
                if 'HISTORY' in header:
                    header.insert('HISTORY', card)
                else:
                    header.append(card)
            elif value != header[name]:
                if type(header[name]) == str:
                    header[name] = 'MULTI'
                elif name not in req_keys:
                    header.remove(name)

    return header




from .base import _ValidHDU, _OIFITS1HDU, _OIFITS2HDU
from .. import utils as _u

from astropy.io import fits as _fits

class _PrimaryHDU(
        _ValidHDU,
        _fits.PrimaryHDU
      ):
    def __repr__(self):
        return f"<{type(self).__name__} at {hex(id(self))} {self._diminfo()}>"
    def __str__(self):
        return f"<{type(self).__name__} {self._diminfo()}>"
    def _diminfo(self):
        if self.data is None:
            return "(void)"
        return f"({'×'.join(self.data.shape)})"

    @classmethod
    def match_header(cls, header):
        cards = getattr(cls, '_CARDS', None)
        if cards is None or 'CONTENT' not in cards['name']:
            return NotImplementedError
        version = cards[cards['name'] == 'CONTENT']['default'][0]
        return (_fits.PrimaryHDU.match_header(header) and
                header.get('CONTENT', 'OIFITS1') == version)

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        if hasattr(cls, '_CARDS') and 'CONTENT' in cls._CARDS['name']:
            _fits.hdu.base._BaseHDU.register_hdu(cls)

    def to_version(self, n):
        classes = {1: PrimaryHDU1, 2: PrimaryHDU2}
        cls = type(self)
        newcls = classes[n]
        if cls == newcls:
            return self
        hdu = newcls(data=self.data, header=self.header)
        hdu.header['CONTENT'] = f"oifits{n}"
        return hdu
    
    def __init__(self, data=None, header=_fits.Header(), keywords={}):
        
        keywords = {k.upper(): v for k, v in keywords.items() if v is not None}

        if 'content' in keywords:
            del keywords['content']
    
        for card in self._CARDS:
            name = card['name']
            comment = card['comment']
            if name in keywords:
                value = keywords[name]
                if isinstance(value, (list, tuple)):
                    value, comment = value[0:2]
                header.set(name, value, comment)
                del keywords[name]
            elif card['required'] and name not in header:
                value = card['default']
                if value is not None:
                    header.set(name, value, comment)
        
        for name, value in keywords.items():
            if not isinstance(value, (tuple, list)):
                value = [value]
            header.set(name, *value)
        
        super().__init__(data=data, header=header)
                


class PrimaryHDU1(
        _PrimaryHDU,
        _OIFITS1HDU,
      ):
    _CARDS = [
        ('CONTENT', False, _u.is_oifits1, 'OIFITS1', 'format by Pauls et al. (2005), PASP 117,1125'),
    ]

        

class PrimaryHDU2(
        _PrimaryHDU,
        _OIFITS2HDU,
      ):
    _CARDS = [
        ('CONTENT',  True,  _u.is_oifits2,       'OIFITS2', 'format by Duvert et al. (2017), A&A 597, A8'),
        ('TELESCOP', True,  _u.is_nonempty,      'N/A', 'name of the telescope array'),
        ('INSTRUME', True,  _u.is_nonempty,      'N/A', 'name of the interferometric instrument'),
        ('OBJECT',   True,  _u.is_nonempty,      'N/A', 'astronomical object'),
        ('REFERENC', False, _u.is_nonempty,      'N/A', 'bibliographic reference'),
        ('PROG_ID',  False, _u.is_nonempty,      'N/A', 'observing programme ID'),
        ('PROCSOFT', False, _u.is_nonempty,      'N/A', 'data processing software'),
        ('OBSTECH',  False, _u.is_nonempty,      'N/A', 'beam recombination technique'),
        ('RA',       False, _u.is_num,           None, 'right ascension (deg)'),
        ('DEC',      False, _u.is_num,           None, 'declination (deg)'),
        ('EQUINOX',  False, _u.is_num,           2000., 'equinox (yr)'),
        ('RADECSYS', False, _u.is_nonempty,      'ICRS', 'celestial coordinate frame'),
        ('SPECSYS',  False, _u.is_nonempty,      'N/A', ''),
        ('TEXPTIME', False, _u.is_strictpos,     None, 'total exposure time'),
        ('MJD-OBS',  False, _u.is_num,           None, 'MJD at start of observation (d)'),
        ('MJD-END',  False, _u.is_num,           None, 'MJD at end of observation (d)'),
        ('BASE_MIN', False, _u.is_strictposnum,  None, 'minimum baseline length (m)'),
        ('BASE_MAX', False, _u.is_strictposnum,  None, 'maximum baseline length (m)'),
        ('WAVELMIN', False, _u.is_strictposnum,  None, 'minimum wavelength (m)'),
        ('WAVELMAX', False, _u.is_strictposnum,  None, 'maximum wavelength (m)'),
        ('NUM_CHAN', False, _u.is_strictposint,  None, 'number of spectral channels'),
        ('SPEC_RES', False, _u.is_strictposnum,  None, 'spectral resolution'),
        ('VIS2ERR',  False, _u.is_strictposnum,  None, 'typical uncertainty in squared visibilities'),
        ('VISPHERR', False, _u.is_strictposnum,  None, 'typical uncertainty in phases (deg)'),
        ('T3PHIERR', False, _u.is_strictposnum,  None, 'typical uncertainty in closure phases (deg)'),
    ]



"""
Defines the FITS Header Data Units (HDUs) of the OIFITS format: primary
header, binary tables containing interferometric data and instrumental 
properties.
"""

from .array import *
from .wavelength import *
from .target import *
from .vis import *
from .vis2 import *
from .t3 import *
from .flux import *
from .corr import *
from .inspol import *
from .primary import *



from astropy.io import fits as _fits
from astropy import table as _table
from astropy.time import Time as _Time
from numpy import ma as _ma
import numpy as _np
import scipy.sparse as _sparse
import re as _re 

from matplotlib import pylab as _plt

from .hdu import *
from . import utils as _u
from . import fitsutils as _fu

from .hdu.base import _ValidHDU
from .hdu.table import _OITableHDU
from .hdu.data import _DataHDU
from .hdu.target import _TargetHDU
from .hdu.array import _ArrayHDU
from .hdu.t3 import _T3HDU
from .hdu.vis2 import _Vis2HDU
from .hdu.vis import _VisHDU
from .hdu.flux import _FluxHDU
from .hdu.wavelength import _WavelengthHDU
from .hdu.referenced import _Referenced
from .hdu.corr import _CorrHDU
from .hdu.inspol import _InspolHDU
from .hdu.primary import _PrimaryHDU


def open(filename, mode='readonly', lazy_load_hdus=True, **kwargs):
    """

Open an OIFITS file.

Arguments
---------

filename (str)
    File name
mode (str, optional, default: 'readonly')
    Read mode
lazy_load_hdus (bool, optional, default: True)
    Whether to only load HDUs if needed

    """
    # we need to read the primary HDU to see which FITS version it is
    if mode == 'ostream':
        cls = OIFITS2
    else:
        with _fits.open(filename, lazy_load_hdus=True) as hdulist:
            if list.__len__(hdulist): # len(h) would load all HDUs
                content = hdulist[0].header.get('CONTENT', '')
                cls = OIFITS2 if content == 'OIFITS2' else OIFITS1
            else:
                cls = OIFITS2
    hdus = cls.fromfile(filename, lazy_load_hdus=lazy_load_hdus, **kwargs)
    return hdus

def openlist(filenames):
    """

Open a list of OIFITS files and merge them.

Arguments
---------

filenames (str × N)
    File names

    """
    hdulists = [open(f, lazy_load_hdus=False) for f in filenames]
    hdulist = _merge(*hdulists, _inplace=True)

    return hdulist

def merge(*hdulists):
    """

Merge several OIFITS.

Arguments
---------

hdulists1, hdulist2, ...
    OIFITS objects 

    """
    return _merge(*hdulists, _inplace=False)     

def _merge(*hdulists, _inplace=False):


    # Use the latest OIFITS version used by the OIFITS to be merged
    # We order them by newest version first.

    oiver = [getattr(hdulist, '_OI_VER', 0) for hdulist in hdulists]
    order = _np.argsort(oiver)[::-1]
    hdulists = [hdulists[o] for o in order]
    
    
    maxver = oiver[order[0]]
    cls = type(hdulists[order[0]])

    # Copy files if necessary
    if not _inplace:
        hdulists = [hdulist.copy() for hdulist in hdulists]
    
    # Trick to avoid Delayed columns
    # for hdulist in hdulists:
    #    for hdu in hdulist:
    #        nope = hdu.data

    # A flat array containing all HDUs.  They keep knowledge of their
    # container.
    hdus = [hdu for hdulist in hdulists for hdu in hdulist]

    
    # The processing steps:
    # * build a composite header from OIFITS primary headers
    # * convert non-void primary headers into image HDUs
    _process_primaryHDUs(hdus)

    # * remove duplicates (OI tables only)
    _remove_equal_OITableHDUs(hdus)

    # * merge TargetHDU, ArrayHDU when possible
    _merge_OITableHDUs(hdus, cls=(_TargetHDU, _ArrayHDU))

    # * duplicate references ARRNAME, INSNAME, CORRNAME leads
    #   to renaming.  Unfortunately HDUs to be merged will be
    #   needlessly copied here
    _rename_conflicting_OITableHDUs(hdus)

    # * merge interferometric data tables where possible
    _merge_OITableHDUs(hdus, cls=_DataHDU)

    # * if some extensions are of the wrong version, convert them
    for i, hdu in enumerate(hdus):
        if isinstance(hdu, _ValidHDU):
            hdus[i] = hdu.to_version(maxver)
    
    # * update the primary HDU header inferring some keywords from 
    #   contents
    merged = cls(hdus)
    merged.update_primary_header()
    
    # * final tweaks
    merged.verify('silentfix+ignore')
    merged.sort()
    merged.update_extver()

    return merged

def _process_primaryHDUs(hdus):
    
    # merge primary headers of OIFITs primary HDUs 
    headers = [hdu.header for hdu in hdus if isinstance(hdu, _PrimaryHDU)]
    header = _fu.merge_fits_headers(*headers)
    hdus[0].header = header
    # delete void OIFIT Primary headers and convert others to images
    for i in range(len(hdus)-1, 0, -1):
        hdu = hdus[i]
        if isinstance(hdu, _fits.PrimaryHDU):
            if hdu.data is not None:
                hdus[i] = _fits.ImageHDU(hdu)
            else:
                del hdus[i]
       
def _remove_equal_OITableHDUs(hdus, cls=_OITableHDU):
    
    for i, hdu in enumerate(hdus):

        if not isinstance(hdu, cls):
            continue

        for j in range(len(hdus)-1, i, -1):
            if hdus[j] == hdu:
                del hdus[j]     


def _rename_conflicting_OITableHDUs(hdus, cls=_Referenced):

    for i, hdu1 in enumerate(hdus):

        if not isinstance(hdu1, cls):
            continue
            
        refkey = getattr(hdu1, '_REFERENCE_KEY', None)
        if not refkey:
            continue 

        # We then look if any further HDU has a duplicate name of hdus[i]. 
        # Sequential Name behaves like a string 'refname_number' aware
        # of its own numbering.

        hdus2 = hdus[i + 1:]
        
        refname1 = _u.SequentialName(hdu1.header[refkey])
        refhdus2 = [hdu2 for hdu2 in hdus2 if hdu2 & hdu1]
        refnames2 = [_u.SequentialName(h.header[refkey]) for h in refhdus2] 
        equal = [refname1 == refname2 for refname2 in refnames2]
        indices = _np.argwhere(equal)[:,0]
        nequal = len(indices)
        if not nequal:
            continue

        # If there are equal names, then rename using name_number.  We
        # need to look at name_number to pick a non used number...
        # seqname1 % seqname2 means same name but different number.  From
        # the list, we can pick available names. 
     
        refs = [refname2 for refname2 in refnames2 if refname2 % refname1] 
        new_refnames2 = refname1.next_available(refs, nequal)

        for index in indices:

            #refname2 = refnames2[index]
            refhdu2 = refhdus2[index]
            new_refname2 = str(new_refnames2.pop(0))
            refhdu2.rename(new_refname2)

            #container = refhdu2.get_container()
            #if not container:
            #    continue
        
            #referrers = container.get_referrers(refhdu2)
            #for h in referrers:
            #    h.header[refkey] = new_refname2
            #
            #refhdu2.header[refkey] = new_refname2
            #
            #if refkey == 'INSNAME':
            #    for h in container.getInspolHDU():
            #        insname = h.data['INSNAME'] 
            #        h.data['INSNAME'][insname == refname2] = new_refname2


def _merge_OITableHDUs(hdus, cls=_OITableHDU):

    # find all equal HDUs and remove them

    for i, hdu in enumerate(hdus):    

        if not isinstance(hdu, cls):
            continue
        
        # Find all mergeable HDUs, merge them 
        nhdu = len(hdus)
        is_mergeable = [hdus[k] % hdu if k > i else False for k in range(nhdu)]
        where_mergeable = _np.argwhere(is_mergeable)[:,0]
        if len(where_mergeable):
            mergeable = [hdus[m] for m in where_mergeable]
            hdus[i] = hdus[i].merge(*mergeable)
            for j in where_mergeable[::-1]:
                del hdus[j]
   
class _OIFITS(_fits.HDUList):

    _merge_station_distance = 0.1    # 0.1 m
    _merge_array_distance = 10       # 10 m
    _merge_target_distance = 2.5e-8  # ~0.005 mas 
    _merge_target_name_match = False # allow different target designations.

    @staticmethod
    def set_merge_settings(*, station_distance=0.1, array_distance=10, 
           target_distance=2.5e-8, target_name_match=False):
        """

Set the default behaviour when merging several OIFITS objects

Arguments
---------

station_distance (float, default: 0.1)
    Maximum distance (m) for stations to be considered the same
array_distance (float, default: 10)
    Maxiumum distance (m) for array centres to be considered from
    the same array
target_distance (float, default: 2.5e-8 i.e. 5 mas)
    Maximum distance (deg) for two targets to be considered the
    same one
target_name_match (bool, default: False)
    Whether target names must match exactly to be considered the
    same one

        """
        _OIFITS._merge_station_distance = station_distance
        _OIFITS._merge_array_distance = array_distance
        _OIFITS._merge_target_distance = target_distance
        _OIFITS._merge_target_name_match = target_name_match

    def __repr__(self):

        name = type(self).__name__
        str_ = [str(h) for h  in self]
        return f"<{name} at {hex(id(self))}: {' '.join(str_)}>"

    def __str__(self):

        name = type(self).__name__
        str_ = [str(h) for h  in self]

        return f"<{name}: {' '.join(str_)}>"
   
    def __init__(self, hdus=[], file=None):
        super().__init__(hdus=hdus, file=file)  
        for hdu in list.__iter__(hdus):
            hdu._container = self

    # original _read_next_hdu() uses super().append(), ruining any clean 
    # attempt to subclass HDUList
    def _read_next_hdu(self):
        
        has_new_hdu = super()._read_next_hdu()
        if has_new_hdu:
            last_index = list.__len__(self) - 1 # len(x) would load all HDUs
            hdu = self[last_index]
            if isinstance(hdu, _OITableHDU):
                hdu._container = self
        return has_new_hdu

    def get_version(self):
        return self._OI_VER

    def _verify(self, option='warn'):

        errors = super()._verify(option) 
      
        clsname = type(self).__name__
 
        # check OI extensions are valid names and fit the OIFITS version
        for hdu in self[1:]:
            
            extname = hdu.header.get('EXTNAME', '')
            extrevn = hdu.header.get('OI_REVN', 0)
            
            if extname[0:3] == 'OI_' and not isinstance(hdu, _OITableHDU):
                err_text = f"Invalid OIFITS extention: {extname}"
                fix_text = "Replaced underscore by dash"
                def fix(hdu=hdu):
                    hdu.header['EXTNAME'] = 'OI-' + extname[3:]
                err = self.run_option(option, err_text=err_text,
                                  fix_text=fix_text, fix=fix)
                errors.append(err)
              
            if hdu._OI_VER != self._OI_VER: 
                err_text = f"Extension {extname} rev. {extrevn} in {clsname}"
                err = self.run_option(option, err_text=err_text, fixable=False) 
                errors.append(err)
       
        # silently fix EXTVER 
        self.update_extver()
        self.update_primary_header()

        return errors

    def update_uv(self):
        """

Update the (u, v) coordinates (UCOORD, VCOORD) using information of
the array and target information contained in OI_ARRAY and OI_TARGET 
tables.

It is a low-precision routine not meant for high precision work.

Warnings
--------

(u, v) may differ from (UCOORD, VCOORD) determined by a data processing
software because

a. (UCOORD, VCOORD) can be averaged independently from MJD 
(see OIFITS standard)
b. Atmospheric refraction is dealt with approximately, while (UCOORD,
   VCOORD) may have none to full modelling of the atmosphere.

For the VLTI, the difference amount to 0.1-0.2% error on baselines
(a few centimetres).

        """
        for hdu in self.get_dataHDUs():
            hdu.update_uv()

    def get_HDUs(self, exttype, filter=None):
        """
Get all HDUs of a given extension type matching given criteria

Arguments
---------

exttype (type)
    Extension type
filter (func)
    Function taking an extension object and returning either True or 
    False

        """
        hdus = [h for h in self[1:] if isinstance(h, exttype)]
        
        if filter is not None:
            hdus = [h for h in hdus if filter(h)]

        return hdus
    
    def get_dataHDUs(self, /, filter=None):
        """
Get all HDUs containing optical interferometry data.  

Keyword arguments

        """
        return self.get_HDUs(_DataHDU, filter)

    def get_HDU(self, extype, /, filter=None):
        """
Get the first HDU of a given extension type matching given criteria

Arguments
---------

exttype (type)
    Extension type
filter (func)
    Function taking an extension object and returning either True or 
    False

        """
        hdus = self.get_HDUs(extype, filter=filter)
        
        if not hdus:
            return None
        return hdus[0]

    def get_OITableHDUs(self):
        """
Get all HDUs containing an OI binary table
        """
        return self.get_HDUs(_OITableHDU)

    def get_arrayHDUs(self):
        return self.get_HDUs(_ArrayHDU)
    
    def get_arrayHDU(self, arrname):
        """
Get HDU containing array description (OI_ARRAY)

Arguments
---------

arrname (str)
    Name of the array

        """
        def same_arrname(h): return h.get_arrname() == arrname
        return self.get_HDU(_ArrayHDU, same_arrname)
    
    def get_vis2HDUs(self):
        """
Get the HDUs containing square visibility amplitude data
        """
        return self.get_HDUs(_Vis2HDU)
    
    def get_visHDUs(self):
        """
Get the HDUs containing square visibility data
        """
        return self.get_HDUs(_VisHDU)
    
    def get_t3HDUs(self):
        """
Get the HDUs containing closure (3T) data
        """
        return self.get_HDUs(_T3HDU)

    def get_targetHDU(self):
        """
Get the HDU containing target information (OI_TARGET)
        """
        return self.get_HDU(_TargetHDU) 

    def get_fluxHDUs(self):
        return self.get_HDUs(_FluxHDU)

    def get_wavelengthHDUs(self):
        """
Get all HDUs containing wavelength information (OI_WAVELENGTH)
        """
        return self.get_HDUs(_WavelengthHDU)

    def get_wavelengthHDU(self, insname):
        """
Get the HDU containing wavelength information (OI_WAVELENGTH) of a 
given instrumental setup.

Arguments
---------

insname (str)
    Name of the instrumental setup

        """ 
        def same_insname(h): return h.get_insname() == insname
        return self.get_HDU(_WavelengthHDU, same_insname)

    def get_corrHDUs(self):
        """
Get all HDUs containing correlation information (OI_CORR)
        """
        return self.get_HDUs(_CorrHDU)

    def get_corrHDU(self, corrname):
        """
Get the HDU containing a correlation matrix (OI_CORR)

Arguments
---------

corrname:
    Name of the correlation matrix

        """
        def same_corrname(h): h.get_corrname() == corrname
        return self.get_HDU(_CorrHDU, same_corrname)

    def get_inspolHDUs(self):

        return self.get_HDUs(_InspolHDU)

    def get_inspolHDU(self, arrname):
        """
Get the HDU containing the instrumental polarisation (OI_INSPOL)
corresponding to a given array.

Arguments
---------

arrname:
    Name of the array

        """
        def same_arrname(h): return h.get_arrname() == arrname
        return self.get_HDU(_InspolHDU, same_arrname)

    def _to_table(self, *, correlations=None, remove_masked=False,
        **kwargs):

        dataHDUs = self.get_dataHDUs()

        return_corr = correlations is not None

        # join all tables generated by each data HDU
        tabs = [h._to_table(full_uv=True, correlations=return_corr,
                    remove_masked=remove_masked, **kwargs) for h in dataHDUs]
        colnames = tabs[0].colnames 
        cols = [_ma.hstack([t[n] for t in tabs]) for n in colnames]
        tab = _table.Table(cols, names=colnames)
        for name in colnames:
            tab.columns[name].format = tabs[0].columns[name].format
        
        if not return_corr:
            return tab

        corr = _sparse.identity(len(tab), format='dok') 

        # treat each OI_CORR separately, then look up indices
        # in the full table
        for corrHDU in np.unique(self.get_corrHDUs()):

            corrname = corrHDU.get_corrname()
            keep = tab['CORRNAME'] == corrname
            
            # global index: goes from 0 to len(tab) - 1
            # local index: CORRINDX tabulated for each OI_CORR
            gi = np.argwhere(keep)[:,0]
            li = tab['CORRINDX'][keep]

            xmatch = {l: g for l, g in zip(li, gi) if not li.mask}

            # looks quite pedestrian, can't I vectorise that?
            for line in corrHDU.data:
                i = xmatch[line['IINDX']]
                j = xmatch[line['JINDX']]
                val = line['CORR']
                if val != 0:
                    corr[i,j] = val
                    corr[j,i] = val

        tab.remove_columns(['CORRNAME', 'CORRINDX'])

        if correlations == 'csr':
            corr = corr.tocsr()
        elif correlations == 'csc':
            corr = corr.tocsc()
        elif correlations == 'coo':
            corr = corr.tocoo()
        elif correlations == 'numpy':
            corr = corr.toarray()
        elif correlations == 'matrix':
            corr = corr.todense()
        elif correlations == 'dok':
            pass
        else:
            raise ValueError(f"wrong correlation matrix format: {correlations}")

        return tab, corr

    def append(self, h2):
        """Not implemented"""
        raise NotImplementedError()

    def pop(self, index=-1):
        """Not implemented"""
        raise NotImplementedError()

    def copy(self):
        """

Create a duplicate of an OIFITS, without an attached file, with data 
and header are copied

        """
        return type(self)([h.copy() for h in self])

    def __add__(self, other):
        return merge(self, other)

    def bin_spectral_channels(self, R):
        """

Bin spectral channels down to a given spectral resolution.  HDUs with lower 
resolution are left untouched.   

If nchan is the number of resulting spectral channels, the resolution at the
middle of the band will be between R and R * (1 + 1 / nchan)

Argument
--------

R (float)
    Desired spectral resolution.  

Return value
------------

An OIFITS1 or OIFITS2 object.

        """
        if any(isinstance(h, (_CorrHDU, _InspolHDU)) for h in self):
           msg = 'Binning OIFITS with correlation or polarimetry info'
           raise NotImplementedError(msg)

        hdulist = [h.copy() for h in self 
                        if not isinstance(h, (_DataHDU, _WavelengthHDU))]

        for whdu in self.get_wavelengthHDUs():

            ins = whdu.get_insname()
            whdu, weights = whdu._bin_helper(R)
            
            dhdus = self.get_dataHDUs(filter=lambda h: h.get_insname() == ins)
            dhdus = [dhdu._bin_helper(weights) for dhdu in dhdus]

            hdulist += [whdu, *dhdus]

        return type(self)(hdulist)

    def update_extver(self):
        """

Update the EXTVER header keyword in extensions sharing the same name. 
While certainly unused by applications, this is a requirement of the
FITS standard.

        """
        extnames = _np.unique(h.header.get('EXTNAME', None) for h in self[1:])
        for extname in extnames:
            hdus = [h for h in self[1:] 
                            if h.header.get('EXTNAME', None) == extname]
            if len(hdus) > 1:
                for i, h in enumerate(hdus):
                    h.header['EXTVER'] = i + 1

    def to_version(self, n):
        """
Convert an OIFITS object to version 1 or 2 of the standard.

Arguments
---------

n (int: 1 or 2)
    Version of the OIFITS standard

        """
        cls = type(self)
        for newcls in type(self).__base__.__subclasses__():
            if newcls._OI_VER == n:
                break

        hdulist = newcls([hdu.to_version(n) for hdu in self])

        return hdulist 

    def update_primary_header(self):
        """

Update the primary header to match the information in OI table 
extensions

        """ 
        header = self[0].header
        datahdus = self.get_dataHDUs()
        
        # fix dates
        mjdobs = min(h.MJD.min() for h in datahdus)
        header['DATE-OBS'] = _Time(mjdobs, format='mjd').isot 
        
        # deal with keywords for atomic observations
        targets = self.get_targetHDU().data
        if len(targets) == 1:
            wavehdus = self.get_wavelengthHDUs()
            mjdend = max(h.MJD.max() for h in datahdus)
            header['OBJECT'] = targets['TARGET'][0] 
            header['RA'] = targets['RAEP0'][0]
            header['DEC'] = targets['DECEP0'][0]
            header['EQUINOX'] = targets['EQUINOX'][0]
            header['MJD-OBS'] = mjdobs
            header['MJD-END'] = mjdend
            wavelmin = min(h.EFF_WAVE.min() for h in wavehdus)
            header['WAVELMIN'] = float(f"{wavelmin:10.4g}")
            wavelmax = max(h.EFF_WAVE.max() for h in wavehdus)
            header['WAVELMAX'] = float(f"{wavelmax:10.4g}")
        else:
            for keyw in ['RA', 'DEC', 'UTC', 'LST', 'EQUINOX', 'RADECSYS', 
                'TEXPTIME', 'MJD-OBS', 'MJD-END', 'BASE_MIN', 'BASE_MAX', 
                'WAVELMIN', 'WAVELMAX', 'NUM_CHAN', 'VIS2ERR', 'VISPHERR', 
                'T3PHIERR']:
                if keyw in header:
                    del header[keyw]

        # Deal with all MULTI keywords
        if len(targets) > 1:
            header['OBJECT'] = 'MULTI'

        arrnames = _np.unique([h.get_arrname() for h in datahdus])
        if len(arrnames) == 1:
            header['TELESCOP'] = arrnames[0]
        else:
            header['TELESCOP'] = 'MULTI'

        insnames = _np.unique([h.get_insname() for h in datahdus])
        if len(insnames) == 1:
            header['INSMODE'] = insnames[0]
        else:
            header['INSMODE'] = 'MULTI'

            # deduce instrument
        ins = [_re.sub('([A-Za-z]+).*', '\\1', i).upper() for i in insnames]
        ins = _np.unique(ins)
        if len(ins) == 1:
            ins = ins[0]
        else:
            ins = 'MULTI'
        self[0].header['INSTRUME'] = ins

        # missing ones
        for keyw in ['PROG_ID', 'REFERENC', 'PROCSOFT', 'OBSTECH',
                     'OBSERVER', 'TELESCOP']:
            if keyw not in header:
                header[keyw] = 'UNKNOWN'

    def visualize(self, xvar, observable, /, *, 
            group_by=None, color_by=None, fig=None, **kwargs):
        """
Quick visualisation of an interferometric observable as a function of
time or baseline.  

Arguments
---------

xvar (str)
    x variable of the plot. It can be 'MJD', 'date', or 'time', 'baseline', 
    or 'spatial_frequency'.

observable (str)
    y variable. It can be any tabulated observable of the standard like
    VIS2DATA, T3AMP, FLUXDATA (OIFITS2), etc.

Keyword arguments
-----------------

fig (int or matplotlib.figure.Figure, optional)
    Matplotlib figure or figure number.  

group_by (str, optional)
    Subplots are grouped by 'target' or 'sta_config'.

color_by (str, optional)
    Data points are coloured according to 'target', 'sta_config', or
    'insname'.

target  (default: all are kept)
    Target name or list of target names

insname  (default: all are kept)
    Instrument configuration name or list thereof

arrname  (default: all are kept)
    Array name or list of array names

mjd_min  (default: all are kept)
    Minimum Modified Julian date

mjd_max  (default: all are kept)
    Maximum Modified Julian date

wavelmin (default: 0)
    Minimum wavelength

wavelmax (default: +inf)
    Maximum wavelength


Returns
-------

fig (matplotlib.figure.Figure)
    A matplotlib figure.

        """
        res = self._plot_helper(xvar, observable, **kwargs)
        (target, inscfg, stacfg, mjd), x, (y, dy) = res

        if xvar in  ['MJD', 'mjd']:
            x = x.to_value('mjd')
        elif xvar in ['date', 'time']:
            x = x.to_datetime()

        if group_by is None:
            if xvar in ['MJD', 'mjd', 'date', 'time']:
                group_by = 'sta_config'
            else:
                group_by = 'target'

        if group_by == 'target':
            key = target
            subkeys = [stacfg, inscfg]
        else:
            key = stacfg
            subkeys = [target, inscfg]

        if color_by == 'target':
            subkey = target
        elif color_by == 'sta_config':
            subkey = stacfg
        elif color_by == 'insname':
            subkey = inscfg
        elif color_by == 'mjd':
            subkey = mjd
        else:
            if len(_np.unique(subkeys[0])) > 1:
                subkey = subkeys[0]
            else:
                subkey = subkeys[1]

        unique_key, unique_index = _np.unique(key, return_inverse=True)
        unique_subkey, unique_subindex = _np.unique(subkey, return_inverse=True)
        
        if fig is None or isinstance(fig, int):
            fig = _plt.figure(fig)
            fig.clf()
        
        naxes = len(unique_key)
        ny = int(_np.sqrt(2 * naxes))
        nx = int(_np.ceil(naxes / ny))
        axes = fig.subplots(ny, nx, sharex=True, sharey=True, squeeze=False)
        
        fig.subplots_adjust(hspace=0, wspace=0)
        for j in range(0, ny):
            for i in range(0, nx):
                if j < nx - 1:
                    axes[j][i].set_xticklabels([])
                else:
                    axes[j][i].set_xlabel(xvar)
                if i > 0:
                    axes[j][i].set_yticklabels([]) 
                else:
                    axes[j][i].set_ylabel(observable)

        for i in range(len(unique_key)):
            k = unique_key[i]
            keep = unique_index == i
            ax = fig.axes[i]
            xi, yi, dyi = x[keep], y[keep], dy[keep]
            subindexi = unique_subindex[keep]
            for j in range(len(unique_subkey)):
                subk = unique_subkey[j]
                keep = subindexi == j
                xj, yj, dyj = xi[keep], yi[keep], dyi[keep]
                ax.errorbar(xj, yj, yerr=dyj, label=subk, linestyle='none')
            ax.text(0.50, 0.99, k, 
                ha='center', va='top', transform=ax.transAxes)
        ax.legend()

        return fig
 
    # this function returns the data to plot:
    # * x variable (time, projected baseline length, or spatial frequency)
    # * y variable (value and error of the observable)
    # * target
    # * baseline used in the format 'ARRAY/STATION1-...'
    def _plot_helper(self, xvar, observable,  /, **kwargs):

        tab = self.to_table(remove_masked=True, observable=observable, **kwargs)

        if not tab:
            raise RuntimeError('no data match criteria')

        if xvar in ['MJD', 'date', 'time']:
            x = _Time(tab['MJD'], format='mjd')
        elif xvar in ['baseline', 'spatial_frequency']:
            x = _np.zeros_like(tab['MJD'])
            t3 = ~tab['U2COORD'].mask
            t2 = ~tab['U1COORD'].mask & ~t3 
            u1, v1 = tab['U1COORD'][t2], tab['V1COORD'][t2]
            b1 = _np.sqrt(u1 ** 2 + v1 ** 2)
            x[t2] = b1
            u1, v1 = tab['U1COORD'][t3], tab['V1COORD'][t3]
            u2, v2 = tab['U2COORD'][t3], tab['V2COORD'][t3]
            u3, v3 = -u1-u2, -v1-v2
            b1 = _np.sqrt(u1 ** 2 + v1 ** 2)
            b2 = _np.sqrt(u2 ** 2 + v2 ** 2)
            b3 = _np.sqrt(u3 ** 2 + v3 ** 2)
            x[t3] = abs(b1 * b2 * b3) ** (1/3)
            if xvar == 'spatial_frequency':
                x /= tab['EFF_WAVE']
        else:
            choices = 'MJD, date, time, baseline, or spatial_frequency'
            raise NotImplementedError(f"x variable must be {choices}")

        if not isinstance(x, _Time):
            x = _np.asarray(x)
        y = _np.asarray(tab['value'])
        dy = _np.asarray(tab['error'])
        
        arr = tab['ARRNAME'].tolist()
        cfg = tab['STA_CONFIG'].tolist()        
        stacfg = _np.asarray([f"{a}/{b}" for a, b in zip(arr, cfg)])
        inscfg = _np.asarray(tab['INSNAME'])        
        mjd = _np.asarray(tab['MJD'])

        target = _np.asarray(tab['TARGET'])

        return (target, inscfg, stacfg, mjd), x, (y, dy) 
 

class OIFITS1(_OIFITS):
    """Top-level class of Optical Interferometry FITS format, version 1."""
    _OI_VER = 1

    def to_table(self, /, *, remove_masked=False, **kwargs):
        """

Convert to a flat table containing one scalar interferometric 
observable per line.  It is possible to only keep data corresponding
to a given list of targets, observables, instrumental configurations,
array configurations, wavelength range and/or date range. 


Arguments
---------

remove_masked (bool, default: False)
    Remove masked values.

observable (default: all are kept) 
    Observable or list of observables (e.g. VIS2DATA)

observable_type  (default: all are kept)
    Observable type or list of observable types (absolute, differential, etc.)

target  (default: all are kept)
    Target name or list of target names

insname  (default: all are kept)
    Instrument configuration name or list thereof

arrname  (default: all are kept)
    Array name or list of array names

mjd_min  (default: all are kept)
    Minimum Modified Julian date

mjd_max  (default: all are kept)
    Maximum Modified Julian date

wavelmin (default: 0)
    Minimum wavelength

wavelmax (default: +inf)
    Maximum wavelength


Returns
-------

tab (astropy.table.Table)
    A table with one scalar observable per line.

        """
        return self._to_table(remove_masked=remove_masked, **kwargs)


class OIFITS2(_OIFITS):
    """Top-level class of Optical Interferometry FITS format, version 2."""
    _OI_VER = 2
    
    def to_table(self, /, *, correlations=None, remove_masked=False, **kwargs):
        """

Convert to a flat table containing one scalar interferometric 
observable per line. It is possible to only keep data corresponding
to a given list of targets, observables, instrumental configurations,
array configurations, wavelength range, and/or date range. 

Arguments
---------

correlations (bool, default: None)
    Type of correlation matrix
    * None: no matrix is return
    * 'numpy': numpy 2D array
    * 'matrix': numpy matrix
    * 'dok': scipy.sparse matrix in dictionary of keys format
    * 'csc': scipy.sparse matrix in compressed sparse column format
    * 'csr': scipy.sparse matrix in compressed sparse row format
    * 'coo': scipy.sparse matrix in coordinate format

remove_masked (bool, default: False)
    Remove masked values.

observable (default: all are kept) 
    Observable or list of observables (e.g. VIS2DATA)

observable_type  (default: all are kept)
    Observable type or list of observable types (absolute, differential, etc.)

target  (default: all are kept)
    Target name or list of target names

insname  (default: all are kept)
    Instrument configuration name or list thereof

arrname  (default: all are kept)
    Array name or list of array names

mjd_min  (default: all are kept)
    Minimum Modified Julian date

mjd_max  (default: all are kept)
    Maximum Modified Julian date

wavelmin (default: 0)
    Minimum wavelength

wavelmax (default: +inf)
    Maximum wavelength



Returns
-------

tab (astropy.table.Table)
    A table with one scalar observable per line.

corr (scipy.sparse.dok_matrix, optional)
    Sparse correlation matrix in a dictionary of keys format

        """
        return self._to_table(correlations=correlations, 
                            remove_masked=remove_masked, **kwargs)


set_merge_settings = _OIFITS.set_merge_settings



"""
Implementation of the OI_WAVELENGTH binary table extension containing
the instrumental spectral setup.
"""

from .table import _OITableHDU, _OITableHDU11, _OITableHDU22
from .referenced import _Referenced

from .. import utils as _u
import numpy as _np
from astropy.io import fits as _fits


_NW = 'NWAVE'

class _MustHaveWavelengthHDU(_OITableHDU):
    
    _CARDS = [('INSNAME', True, _u.is_nonempty, None, 
        'instrumental setup name for cross-reference')]

    def get_insname(self, shape='none', flatten=False, copy=True):
        x = self.header['INSNAME']
        return self._resize_data(x, shape, flatten, copy)

    def get_wavelengthHDU(self):
        return self._container.get_wavelengthHDU(self.get_insname())

    def _resize_wave_data(self, x, shape='data', flatten=False):

        if self is self.get_wavelengthHDU():
            return x
        
        if shape in ['data', 'table']:
            x = _np.full(self.data_shape(), x)
        if flatten:
            x = x.ravel()

        return x

    def get_nwaves(self):
        
        return len(self.get_wavelengthHDU().data)
        
    def get_wave(self, shape='data', flatten=False):

        wave = self.get_wavelengthHDU().EFF_WAVE
        return self._resize_wave_data(wave, shape, flatten)

    def get_channel(self, shape='data', flatten=False):

        nwave = len(self.get_wavelengthHDU().data)
        channel = _np.arange(1, nwave + 1)
        return self._resize_wave_data(channel, shape, flatten)

    def get_band(self, shape='data', flatten=False):

        band = self.get_wavelengthHDU().EFF_BAND
        return self._resize_wave_data(band, shape, flatten)


class _WavelengthHDU(_MustHaveWavelengthHDU,_Referenced):
    
    _EXTNAME = 'OI_WAVELENGTH'
    _REFERENCE_KEY = 'INSNAME'
    _COLUMNS = [
        ('EFF_WAVE', True, '1E', (), _u.is_strictpos, None, "m",
                                'effective wavelength')]
    
    def _diminfo(self):

        nrows = len(self.data)
        return f"{super()._diminfo()}={nrows}W"

    def is_referred_to_by(self, other):
        return (not isinstance(other, _WavelengthHDU) and
                isinstance(other, _MustHaveWavelengthHDU) and
                self.get_insname() == other.get_insname())

    def __mod__(self, other):

        return False

    def rename(self, new_name):

        old_name = self.header['INSNAME']
        super().rename(new_name)
        
        container = self.get_container()
        if container is None:
            return

        for h in container.get_inspolHDUs():
            to_rename = h.data['INSNAME'] == oldname
            h.data['INSNAME'][to_rename][...] = new_name

    @classmethod
    def from_data(cls, *, insname, version=2, eff_wave, eff_band=0., 
            fits_keywords={}, **columns):

        """

        Build an OI_WAVELENGTH binary table from data

Arguments
---------

    insname (str)
        Name of this table
    version (int)
        OIFITS version if it cannot be deduced from context (optional)

    eff_wave (float, NWAVE)
        Effective wavelength
    eff_band (float, NWAVE)
        Bandwidth (optional, defaults to 0.)

    fits_keywords (dict)
        Additional FITS header keywords (optional)

Additional arguments
--------------------

Any additional keyword argument will be appended as a non-standard FITS 
column with its name prefixed with NS_ 

        """
        fits_keywords = dict(insname=insname, **fits_keywords)
        columns = dict(eff_wave=eff_wave, eff_band=eff_band, **columns)

        return super().from_data(fits_keywords=fits_keywords, **columns)

    def _bin_helper(self, R):

        min, max = _np.minimum, _np.maximum

        # convert to double precision (rounding errors in weights...)
        wave = _np.asarray(self.get_wave(shape='wavelength'), dtype=float)
        band = _np.asarray(self.get_band(shape='wavelength'), dtype=float)

        winf, wsup = wave - band / 2,  wave + band / 2
        wmin, wmax = winf[0], wsup[-1]
        wmed = (wmin + wmax) / 2
        R0 = wmed / ((wmax - wmin) / len(wave))

        if R0 > R and len(wave) > 1:

            nwave = int(_np.ceil(len(wave) * R / R0))

            dw = (wmax - wmin) / nwave
            w1 = winf[0] + dw / 2
            w2 = wsup[-1] - dw / 2
            new_wave = _np.linspace(w1, w2, nwave)
            new_winf = new_wave - dw / 2
            new_wsup = new_wave + dw / 2
            new_band = _np.full((nwave,), dw)

            weights = max(min(wsup, new_wsup[:,None])
                            - max(winf, new_winf[:,None]), 0) / band
            weights = weights.T
            
            oi_colnames = self._get_oi_colnames()
            fmt = f"{nwave}E"

            wave_col = _fits.Column('EFF_WAVE', fmt, 'm', array=new_wave) 
            band_col = _fits.Column('EFF_BAND', fmt, 'm', array=new_band) 
            other_cols = [c for c in self.columns if c.name not in oi_colnames]
            cols = [wave_col, band_col, *other_cols]

            new = self.from_columns(cols, header=self.header)

        else:
            
            new = self.copy()
            weights = None

        return new, weights



class WavelengthHDU1(
        _WavelengthHDU,
        _OITableHDU11,
     ):
    """

First revision of the OI_WAVELENGTH binary table, OIFITS v. 1

    """
    _COLUMNS = [('EFF_BAND', True, '1E', (), _u.is_strictpos, None, "m",
        'effective bandwidth')]


class WavelengthHDU2(
        _WavelengthHDU,
        _OITableHDU22,
      ):
    """

First revision of the OI_WAVELENGTH binary table, OIFITS v. 1

    """
    _COLUMNS = [('EFF_BAND', True, '1E', (), _u.is_pos, None, "m",
        'effective bandwidth')]
  
new_wavelength_hdu = _WavelengthHDU.from_data 



from numpy import isnan, cos, sin
import astropy.units as u
from astropy.coordinates import AltAz

def altaz_frame(location, *, obswl, refraction=False):
    """

Alt-az frame taking into account mean astmospheric refraction.

Arguments
---------

location (astropy.coordinates.EarthLocation)
    Earth location of site 

obswl (astropy.units.Quantity)
    Observing wavelength


Returns
-------

altaz (astropy.coordinates.AltAz)
    Alt-az frame

    """
    
    if not refraction:
        return AltAz(location=location)

    lat = location.lat.to_value("rad")
    h = location.height.to_value("m")

    # Mean temperature at sea level  
    # 
    # Rough approximation to Feulner et al. (2013). On the Origin of 
    # the Surface Air Temperature Difference between the Hemispheres 
    # in Earth's Present-Day Climate. Journal of Climate. 26. 7136-7150.
    # 10.1175/JCLI-D-12-00636.1. 

    T0 = 250 + 50 * cos(lat) # K

    # Mean temperature on site
    #
    # Neumann (1955). Latitudinal variation of tropospheric temperature
    # lapse rate. Arch. Met. Geoph. Biokl. A. 8, 351–353  
    # https://doi.org/10.1007/BF02247093 
    # But we use the 0.0065 value of the Standard US Atmosphere for
    # the low to moderate latitudes

    Lb = 6.5e-3 - 2e-3 * sin(lat) ** 4 # K/m
    T = (T0 - h * Lb - 273.15) # ⁰C 

    # Mean pressure at sea level
    #
    # Variation with latitude is negligible at this level of accuracy
    # (T, P vary easily by 5%).

    P0 = 101325 # Pa

    # Mean pressure on site
    #
    # Let's use the barometric formula for a real atmosphere 
    # (dT/dz < 0, hT = -dz/dT)  to estimate the pressure at 
    # the observatory. It will give an estimate for the 
    # atmospheric refraction needed to determine the apparent
    # position of the source
    
    h0 =  8438 * (T0 / 288.15) # m (isothermal scale height)
    hT = 44330 * (6.25e-3 / Lb) # m (temperature scale height)

    P = P0 * (1 - h/hT) ** (hT/h0) # Pa

    # Relative humidity
    H = 20 * u.percent # % (typical of observatories)

    frame = AltAz(location=location, 
                    pressure=P * u.Pa, 
                    temperature=T * u.deg_C, 
                    obswl=obswl,
                    relative_humidity=H * u.percent) 
    
    return frame

def apply_space_motion(coo, obstime, correct_motion=False):
    """

Update coordinates to current epoch

    """
    # set new epoch
    dt = obstime - coo.obstime
    coo.obstime = obstime
 
    # position correction due to motion (linear approximation)
    pmra = coo.pm_ra_cosdec
    pmde = coo.pm_dec
    rv = coo.radial_velocity
    d = coo.distance
    
    has_pm = ~(isnan(pmra) | isnan(pmde))
    has_rv = ~isnan(rv)

    coo.ra[has_pm] += pmra[has_pm] / cos(coo.dec[has_pm]) * dt[has_pm]
    coo.dec[has_pm] += pmde[has_pm] * dt[has_pm]
    coo.distance[has_rv] += rv[has_rv] * dt[has_rv]

    # motion correction
    # we probably don't need it for OIFITS...
    if correct_motion:
        raise RuntimeError('Not implemented')

    return coo



"""
Implementation of the OI_CORR binary table extension containing corralation
between interferometric data in OI_VIS, OI_VIS2, OI_T3, and OI_FLUX binary
table extensions
"""

from .table import _OITableHDU, _OITableHDU21
from .referenced import _Referenced
from .. import utils as _u
from scipy import sparse as _sparse

import numpy as _np

class _CorrHDUBase(_OITableHDU):
    
    def get_corrname(self, shape='none', flatten=False, default=None):
        corrname = self.header.get('CORRNAME', default)
        return self._resize_data(corrname, shape, flatten)

    def get_corrHDU(self):
        corrname = self.get_corrname()
        if not corrname:
            return None
        return self._container.get_corrHDU(corrname)

class _MayHaveCorrHDU(_CorrHDUBase):
    _CARDS = [('CORRNAME', False, _u.is_nonempty, None, 
        'correlation name for cross-reference')]

class _MustHaveCorrHDU(_CorrHDUBase):
    _CARDS = [('CORRNAME', True,  _u.is_nonempty, None, 
        'correlation name for cross-reference')]

class _CorrHDU(_MustHaveCorrHDU,_Referenced):
    
    _EXTNAME = 'OI_CORR'
    _REFERENCE_KEY = 'CORRNAME'

    _CARDS = [
        ('NDATA', True,  _u.is_strictpos, None, 
            'size of correlation matrix (NDATAxNDATA)'),
    ] 
    _COLUMNS = [
        ('IINDX', True, '1J', (), _u.is_strictpos, None, None,
            '1st index (<= NDATA)'),
        ('JINDX', True, '1J', (), _u.is_strictpos, None, None,
            '2nd index (<= NDATA)'), 
        ('CORR',  True, '1D', (), None,            None, None,
            'correlation C[IINDX,JINDX]')
    ]
    
    def _verify(self, option='warn'):

        errors = super()._verify(option)

        corr_index = self.CORR_INDEX
        if len(_np.unique(corr_index)) == len(coor_index):
            return errors

        err_text = f"Repeated CORR_INDEX in {type(self).__name__}"
        err = self.run_option(option, err_text, fixable=False)
        errors.append(err)

        return errors

    def get_corrname(self):
        return self.header['CORRNAME']

    def __mod__(self, other):

        return False

    def is_referred_to_by(self, other):
        return (not isinstance(other, _CorrHDU) and
                isinstance(other, _CorrHDUBase) and
                self.get_corrname() == other.get_corrname())

    @classmethod
    def from_data(cls, *, corrname, corrmatrix, fits_keywords={}, **columns):

        fits_keywords = dict(corrname=corrname, **fits_keywords)

        # find non-zero elements in the lower triangle, note that
        # indices start at one in OIFITS        
        iindx, jindx, corr = _sparse.find(corrmatrix)
        keep = iindx < jindx
        corr = corr[keep]
        iindx = iindx[keep]
        jindx = jindx[keep]
        iindx += 1
        jindx += 1

        columns = dict(iindx=iindx, jindx=jindx, corr=corr, **columns)        

        super().from_data(fits_keywords=fits_keywords, **columns)
    
class CorrHDU1(
        _CorrHDU,
        _OITableHDU21, # OIFITS2, table rev. 1
    ):
    """

First revision of the OI_CORR binary table, OIFITS v. 2

    """
    pass






from .data import _DataHDU11, _DataHDU22
from .table import _OITableHDU
from .wavelength import _NW

from .. import utils as _u

class _T3HDU(_OITableHDU):
    """
OI_T3 binary table extension containing closure phases and triple 
product amplitudes.
    """
    
    _EXTNAME = 'OI_T3'    
    _COLUMNS = [
        ('STA_INDEX', True, '3I', (3,),   _u.is_strictpos, None, None,
            'station indices in matching OI_ARRAY table'),
        ('U1COORD',   True, '1D', (),     None,            None, "m",
            'u coordinate of first baseline'), 
        ('V1COORD',   True, '1D', (),     None,            None, "m",
            'v coordinate of first baseline'),
        ('U2COORD',   True, '1D', (),     None,            None, "m",
            'u coordinate of second baseline'), 
        ('V2COORD',   True, '1D', (),     None,            None, "m",
            'v coordinate of second baseline'),
        ('T3AMP',     True, 'D',  (_NW,), None,            None, None,
            'Triple amplitude'), 
        ('T3AMPERR',  True, 'D',  (_NW,), None,            None, None,
            'uncertainty on triple amplitude'),
        ('T3PHI',     True, 'D',  (_NW,), None,            None, "deg",
            'closure phase'), 
        ('T3PHIERR',  True, 'D',  (_NW,), None,            None, "deg",
            'uncertainty on closure phase'),
    ]
    
    def get_obs_type(self, name, shape='data', flatten=False):

        type = 'absolute'

        return self._resize_data(type, shape, flatten)

    def get_uvw(self, refraction=False):
        """

Get the (u, v, w) coordinates of the first two baselines

Arguments
---------

refraction (bool, optional, default: False)
    Whether refraction is taken into account

Returns
-------

(u1, v1, w1), (u2, v2, w2)
    Projected baselines in metres

Precision
---------
    A few centimetres for hectometric baselines, mainly due to the uncertainty
    on the time-averaging.
    

        """
        uvw = self.get_stauvw(refraction=refraction)
        uvw1 = uvw[:,1,:] - uvw[:,0,:]
        uvw2 = uvw[:,2,:] - uvw[:,1,:]
      
        return uvw1.T, uvw2.T

    def update_uv(self):

        if self.get_arrayHDU() is None:
            return

        (u1, v1, w1), (u2, v2, w2) = self.get_uvw()
        self.U1COORD = u1
        self.V1COORD = v1
        self.U2COORD = u2
        self.V2COORD = v2
        
    @classmethod
    def from_data(cls, *, u1coord, v1coord, u2coord, v2coord,
            t3phi, fits_keywords={}, **columns):
        """

Build an OI_T3 table from data.  In the following NWAVE indicates the
number of spectral channels in the configuration and NOBS the number
of observations.

Arguments
---------

version (int)
    version of the OIFITS standard if it cannot be determined from
    context (optional)

date (str)
    date at the start of the first observation (optional, deduced from
    mdj and int_time)
insname (str)
    name of the OI_WAVELENGTH table containing the instrumental setup
arrname (str)         
    name of the OI_ARRAY table containing array properties (optional in
    version 1)
fits_keywords (dict)
    additional FITS keywords (optional)

mjd (float, NOBS)
    MJD of observation
int_time (float, NOBS)
    Integration time
target_id (int, NOBS)
    Target ID in the OI_TARGET table
sta_index (int, NOBS × 3)
    Station IDs of the telescope triplet in the OI_ARRAY target (optional 
    in version 1)
u1coord, v1coord, u2coord, v2coord (float, NOBS)
    (u, v) coordinates of the first two baselines.


t3phi (float, NOBS × NWAVE)
    closure phase
t3phierr (float, NOBS × NWAVE)
    uncertainty on the latter (optional, defaults to 0.)
t3amp (float, NOBS × NWAVE)
    triple product amplitude (optional, defaults to NULL) 
t3phierr (float, NOBS × NWAVE)
    uncertainty on the latter (optional, defaults to 0.)
flag (bool, NOBS × NWAVE)
    flag for data (optional, defaults to False)

Arguments in OIFITS1 only
-------------------------

time (float, NOBS)
     Seconds since midnight UT on first observation (DATE-OBS i.e. date)


Arguments in OIFITS2 only
-------------------------

corrindx_t3phi (int, NOBS × NWAVE)
    index in the correlation matrix for the first t3phi (optional)
corrindx_t3amp (int, NOBS × NWAVE)
    index in the correlation matrix for the first t3amp (optional)

Additional arguments
--------------------

Any additional keyword argument will be appended as a non-standard FITS 
column with its name prefixed with NS_ 

        """
        _u.store_default(columns, 't3amp', default=_np.nan)

        return super().from_data(u1coord, v1coord, u2coord, v2coord, t3phi, 
            fits_keywords=fits_keywords, **columns)

class T3HDU1(
        _T3HDU,
        _DataHDU11,
      ):
    """

First revision of the OI_T3 binary table, OIFITS v. 1

    """
    pass
    

class T3HDU2(
        _T3HDU,
        _DataHDU22,
      ):
    """

Second revision of the OI_T3 binary table, OIFITS v. 2

    """
    _COLUMNS = [
        ('CORRINDX_T3AMP', False, '1J', (), _u.is_strictpos, None, None,
            'index of 1st amp. in matching OI_CORR matrix' ), 
        ('CORRINDX_T3PHI', False, '1J', (), _u.is_strictpos, None, None,
            'index of 1st phase in matching OI_CORR matrix'),
    ]

    @classmethod
    def from_data(cls, *, arrname, fits_keywords={}, **columns):

        return super().from_columns(arrname=arrname,
                        fits_keywords=fits_keywords, **columns)

new_t3_hdu = _T3HDU.from_data



"""
Implementation of the OI_FLUX binary table extension containing either
the calibrated fluxes of the targets or the uncalibrated fluxes measured
at the interferometer aperture. 
"""

from .target import _MustHaveTargetHDU
from .array import _MayHaveArrayHDU
from .wavelength import _MustHaveWavelengthHDU, _NW
from .data import _DataHDU, _DataHDU21
from .. import utils as _u


class _FluxHDU(
        _DataHDU,
        _MustHaveTargetHDU,
        _MayHaveArrayHDU,
        _MustHaveWavelengthHDU,
      ):
    _EXTNAME = 'OI_FLUX'
    
    _CARDS = [
        ('CALSTAT', True,  _u.is_calstat, None, 
            'Flux is (C)alibrated or (U)ncalibrated'),
        ('FOV',     False, _u.is_pos,     None, 
            'Field of view (arcsec)'),
        ('FOVTYPE', False, _u.is_fovtype, None, 
            'Type of field of view'),
    ]
    
    _COLUMNS = [
        ('FLUXDATA',          True,  '>f8', (_NW,), None, 'any', None,
            'Flux'),
        ('FLUXERR',           True,  '>f8', (_NW,), None, 'any', None,
            'Flux uncertainty'),
        ('STA_INDEX',         False, '>i2', (),     None, None, None,
            'Station index in matching OI_ARRAY table'),
        ('CORRINDX_FLUXDATA', False, '>i4', (),     None, None, None,
            'Index of 1st flux in matching OI_CORR matrix'),
    ]
    
    
    @classmethod
    def from_data(cls, *, arrname, fluxdata, calibrated=False,
            fov=0., fovtype='FWHM', 
            fits_keywords={}, **columns):
        """
Arguments
---------

version (int)
    version of the OIFITS standard if it cannot be determined from
    context (optional)

date (str)
    date at the start of the first observation (optional, deduced from
    mdj and int_time)
insname (str)
    name of the OI_WAVELENGTH table containing the instrumental setup 
arrname (str)         
    name of the OI_ARRAY table containing array properties (only if
    not calibrated)
corrname (str)
    name of the correlation matrix (only if uncalibrated)

mjd (float, NROWS)
    MJD of observation
int_time (float, NROWS)
    Integration time
target_id (int, NROWS)
    Target ID in the OI_TARGET table
sta_index (int, NROWS)
    station index of the flux (only if uncalibrated)

fluxdata (float, NROWS × NWAVE)
    flux
fluxerr (float, NROWS × NWAVE)
    uncertainty on the latter (optional, defaults to 0.)
flag (bool, NROWS × NWAVE)
    flag (optional, defaults to False)

corrindx_flux_data (int, NROWS)
    index in the correlation matrix for the first fluxdata (optional)

fits_keywords (dict)
    additional FITS keywords (optional)

Additional arguments
--------------------

Any additional keyword argument will be appended as a non-standard FITS 
column with its name prefixed with NS_ 

        """

        shape = _np.shape(fluxdata)
        _u.store_default(columns, 'fluxerr', 0., shape)        

        if calibrated:
            fits_keywords = dict(calstat='C', fov=fov, fovtype=fovtype,
                                 **fits_keywords) 
        else:
            fits_keywords = dict(calstat='U', **fits_keywords)
            
        columns = dict(fluxdata=fluxdata, **columns)

        return super().from_data(arrname=arrname, 
                    fits_keywords=fits_keywords, **columns)
    
    def get_obs_type(self, name, shape='data', flatten=False):
   
        typ = f"{'un' if self.header['CALSTAT'] == 'U' else ''}calibrated flux" 
            
        return self._resize_data(typ, shape, flatten)

    def update_uv(self):
        pass
    
    def _verify(self, option='warn'):

        errors = []
    
        cols = self.columns
        err = None
        if 'FLUXDATA' not in cols.names and 'FLUX' in cols.names:
            err_text = "Column FLUX should be named FLUXDATA"
            fix_text = "Renamed"
            def fix(h=self): h.rename_columns(FLUX='FLUXDATA')
            err = self.run_option(option, err_text, fix_text, fix)

        errors = super()._verify(option)
        if err:
            errors.append(err)

    def __mod__(self, other):

        h1, h2 = self.header, other.header

        return (super().__mod__(other) and
                h1['CALSTAT'] == h2['CALSTAT'] and
                h1.get('FOV', 0) == h2.get('FOV', 0) and
                h1.get('FOVTYPE', '') == h2.get('FOVTYPE', ''))
 

class FluxHDU1(
        _FluxHDU,
        _DataHDU21,
      ):
    """

First revision of the OI_FLUX table, OIFITS v. 2

    """ 
    pass

new_flux_hdu = _FluxHDU.from_data



from .array import _MustHaveArrayHDU
from .table import _OITableHDU, _OITableHDU21
from .wavelength import _NW
from .. import utils as _u

class _InspolHDUBase(_OITableHDU):

    def get_inspolHDU(self):

        if isinstance(self, InspolHDU_v1):
            return self

        arrname = self.header.get('ARRNAME')
        def same_arrname(h): return h.get('ARRNAME', '') == arrname
        hdus = self.get_HDUs(InspolHDU, filter=same_arrname)
        if not hdus:
            return None
        return hdus[0]

    def get_jones_matrix(self, shape='data', flatten=False):

        J = np.zeros((self.data_shape(), 2, 2), dtype=complex)

        inspol = self.get_inspolHDU()

        if inspol is not None:

            insname = self.get_insname()
            poldata = inspol.data

            for i, row in enumerate(self.data):
                match = ((poldata['INSNAME'] == insname)
                        * (poldata['TARGETID'] == row['TARGETID'])
                        * (poldata['MJD'] <= row['MJD'])
                        * (row['MDJ'] <= poldata['MJD-END']))
                if any(match):
                    j = poldata[match][0]
                    J[i,:,0,0] = j['JXX']
                    J[i,:,0,1] = j['JXY']
                    J[i,:,1,0] = j['JYX']
                    J[i,:,1,1] = j['JYY']

        if flatten:
            J = J.reshape((-1, 2, 2))

        return J

class _MayHaveInspolHDU(
        _InspolHDUBase,
        # _MayHaveArrayHDU, -> would give incorrect card
      ):
    pass

class _MustHaveInspolHDU(
        _InspolHDUBase,
        _MustHaveArrayHDU,
      ):
    pass 
    
class _InspolHDU(
        _MustHaveInspolHDU,
      ):
    _EXTNAME = 'OI_INSPOL'
    
class InspolHDU1(
        _InspolHDU, 
        _OITableHDU21, # OIFITS2, table rev. 1
      ):
    """

First revision of the OI_INSPOL binary table, OIFITS v. 2

    """
    _CARDS = [
        ('NPOL',   True, _u.is_strictpos, None, 
            'number of polarisations'),
        ('ORIENT', True, _u.is_nonempty,  None, 
            'orientation of the Jones matrix'),
        ('MODEL',  True, _u.is_nonempty,  None, 
            'method used to determine the Jones matrix')
    ]
    _COLUMNS = [
        ('TARGET_ID',  True, '1I',  (),     _u.is_strictpos, None, None,
            'target ID in matching OI_TARGET table'),
        ('INSNAME',    True, '32A', (),     None,            None, None,
            'name of matching OI_WAVELENGTH table'), 
        ('MJD_OBS',    True, '1D',  (),     None,            None, "d",
            'modified Julian day at start of observation'), 
        ('MJD_END',    True, '1D',  (),     None,            None, "d",
            'modified Julian day at en of observation'),
        ('JXX',        True, 'M',   (_NW,), None,            None, None,
            'Jones matrix element J_xx'), 
        ('JYY',        True, 'M',   (_NW,), None,            None, None,
            'Jones matrix element J_xy'),
        ('JXY',        True, 'M',   (_NW,), None,            None, None,
            'Jones matrix element J_yx'), 
        ('JYX',        True, 'M',   (_NW,), None,            None, None,
            'Jones matrix element J_yy'),
        ('STA_INDEX',  True, '1I',  (),     None,            None, None,
            'station index in matching OI_ARRAY table'),
    ]

    def __mod__(self, other):
        return False

    def get_insname(self, shape='data', flatten=False):

        insname = self.data['INSNAME']
        return self._resize_data(insname, shape, flatten)

    def get_jones_matrix(self, shape='data', flatten=False):

        J = np.empty((nrows, nwaves, 2, 2), dtype=complex)
        J[...,0,0] = poldata['JXX']
        J[...,0,1] = poldata['JXY']
        J[...,1,0] = poldata['JYX']
        J[...,1,1] = poldata['JYY']

        if flatten:
            J = J.reshape((-1, 2, 2))

        return J




from .data import _DataHDU11, _DataHDU22
from .t2 import _T2HDU
from .. import utils as _u
from .wavelength import _NW

import numpy as _np

class _Vis2HDU(_T2HDU):
    """

OI_VIS2 binary table extension containing the squared visibility amplitudes

    """

    _EXTNAME = 'OI_VIS2'
    _COLUMNS = [
        ('VIS2DATA', True, 'D', (_NW,), None, None, None,
            'squared visibility amplitude'), 
        ('VIS2ERR',  True, 'D', (_NW,), None, None, None,
            'uncertaintly on squared visibility amplitude'),
    ]
    
    def get_obs_type(self, name, shape='data', flatten=False):
        
        return self._resize_data('absolute', shape, flatten)

    @classmethod
    def from_data(cls, *, vis2data, fits_keywords={}, **columns):
        """

Build an OI_VIS2 table from data.  In the following NWAVE indicates the
number of spectral channels in the configuration and NOBS the number
of observations.

Arguments
---------

version (int)
    version of the OIFITS standard if it cannot be determined from
    context (optional)

date (str)
    date at the start of the first observation (optional, deduced from
    mdj and int_time)
insname (str)
    name of the OI_WAVELENGTH table containing the instrumental setup
arrname (str)         
    name of the OI_ARRAY table containing array properties (optional in
    version 1)
fits_keywords (dict)
    additional FITS keywords (optional)

mjd (float, NOBS)
    MJD of observation
int_time (float, NOBS)
    Integration time
target_id (int, NOBS)
    Target ID in the OI_TARGET table
sta_index (int, NOBS × 2)
    Station IDs of the telescope pair in the OI_ARRAY target (optional in
    version 1)
ucoord, vcoord (float, NOBS)
    (u, v) coordinates of the  baselines

vis2data (float, NOBS × NWAVE)
    squared visibility amplitudes
vis2err (float, NOBS × NWAVE)
    uncertainty on the latter (optional, defaults to 0.)
flag (bool, NOBS × NWAVE)
    flag for data (optional, defaults to False)

Arguments in OIFITS1 only
-------------------------

time (float, NOBS)
    Seconds since midnight UT on first observation (DATE-OBS i.e. date)

Arguments in OIFITS2 only
-------------------------

corrindx_vis2data (int, NOBS × NWAVE)
    index in the correlation matrix for the first vis2data (optional)

Additional arguments
--------------------

Any additional keyword argument will be appended as a non-standard FITS 
column with its name prefixed with NS_ 

        """
        columns = dict(vis2data=vis2data, **columns)

        return super().from_data(fits_keywords=fits_keywords, **columns)

class Vis2HDU1(
        _Vis2HDU,
        _DataHDU11,
      ):
    """

First revision of the OI_VIS2 binary table, OIFITS v. 1

    """
    pass

class Vis2HDU2(
        _Vis2HDU,
        _DataHDU22,
      ):
    """ 

Second revision of the OI_VIS2 binary table, OIFITS v. 2

    """
    _COLUMNS = [
        ('CORRINDX_VIS2DATA', False, '1J', (), _u.is_strictpos, None, None,
            'index of 1st visib. in matching OI_CORR matrix'),
    ]

new_vis2_hdu = _Vis2HDU.from_data
